import os
from abc import ABC
from typing import Annotated, ClassVar, Literal, Type, TypeAlias, Union

from pydantic import BaseModel, Field
from pydantic.config import ConfigDict
from pydantic.fields import FieldInfo
from typing_extensions import Unpack

from bayesline.api._src.equity.riskmodels_settings import FactorRiskModelSettings
from bayesline.api._src.registry import Settings, SettingsMenu, SettingsTypeMetaData


class PortfolioReportAccessorSettings(Settings):

    axes: dict[str, list[str]]
    metric_cols: list[str]
    pivot_cols: list[str]

    # can be used for multi accessors, one entry for each underlying report
    # e.g. extra_cols = name, risk_model, date # noqa: E800
    # e.g. extra_paths = [["My Setting", "US-All", "2024-12-12"]] # noqa: E800
    extra_paths: list[list[str]] = Field(default_factory=list)
    extra_cols: list[str] = Field(default_factory=list)

    linked_reports: list[int] = Field(default_factory=list)
    tags: list[str] = Field(default_factory=list)


class MeasureSettings(BaseModel, ABC, frozen=True, extra="forbid"):
    """Defines settings for a measure."""

    type: str

    @property
    def columns(self) -> list[str]:
        return [self.type]

    def __init_subclass__(cls, **kwargs: Unpack[ConfigDict]) -> None:
        super().__init_subclass__(**kwargs)
        # check that every field has a default value, so we can instantiate settings
        # without any argument. Fields defined as a: int = 1, a: int = Field(1), and
        # a: list[int] = Field(default_factory=list) are all valid.
        for f in cls.__annotations__:
            msg = f"{f} not does not have a default value"
            if f not in cls.__dict__:
                raise AssertionError(msg)  # fields without ... = Field(...)
            if isinstance(cls.__dict__[f], FieldInfo):
                if cls.__dict__[f].is_required():
                    raise AssertionError(msg)  # with = Field(...)


class PassThroughFactor2DMeasureSettings(MeasureSettings):
    type: Literal["PassThroughFactor2D"] = Field("PassThroughFactor2D", repr=False)
    name: str = Field("PassThroughFactor2D", description="The name of the measure.")

    @property
    def columns(self) -> list[str]:
        return [self.name]


class HoldingsMeasureSettings(MeasureSettings):
    type: Literal["Holdings"] = Field("Holdings", repr=False)

    @property
    def columns(self) -> list[str]:
        return ["Portfolio", "Benchmark", "Active"]


class ExposureMeasureSettings(MeasureSettings):
    type: Literal["Exposure"] = Field("Exposure", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")

    @property
    def columns(self) -> list[str]:
        return ["Portfolio", "Benchmark", "Active"]


class XSigmaRhoMeasureSettings(MeasureSettings):
    type: Literal["XSigmaRho"] = Field("XSigmaRho", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    analytics_space: Literal["absolute", "active", "benchmark"] = Field(
        "absolute",
        description=(
            "Compute the analytics in `absolute` space, `active` space, or compute the "
            "analytics only on the `benchmark`."
        ),
    )

    @property
    def columns(self) -> list[str]:
        return [
            "Exposure (X)",  # X
            "Volatility (\u03c3)",  # σ
            "Correlation (\u03c1)",  # ρ
            "Contribution (X\u03c3\u03c1)",  # Xσρ
        ]


class TimeSeriesXSigmaRhoMeasureSettings(MeasureSettings):
    type: Literal["TimeSeriesXSigmaRho"] = Field("TimeSeriesXSigmaRho", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    analytics_space: Literal["absolute", "active", "benchmark"] = Field(
        "absolute",
        description=(
            "Compute the anlytics in `absolute` space, `active` space, or compute the "
            "analytics only on the `benchmark`."
        ),
    )
    backfill_holdings: bool = Field(
        False,
        description="Backfill the latest holdings, accounting for listing/delisting.",
    )

    @property
    def columns(self) -> list[str]:
        return [
            "Exposure (X)",  # X
            "Volatility (\u03c3)",  # σ
            "Correlation (\u03c1)",  # ρ
            "Contribution (X\u03c3\u03c1)",  # Xσρ
        ]


class TimeSeriesBetaMeasureSettings(MeasureSettings):
    type: Literal["TimeSeriesBeta"] = Field("TimeSeriesBeta", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )

    @property
    def columns(self) -> list[str]:
        return [
            "Beta (\u03b2)",  # β
        ]


class HVaRMeasureSettings(MeasureSettings):
    type: Literal["Historical VaR"] = Field("Historical VaR", repr=False)
    alpha: list[float] = Field([0.1, 0.05])

    @property
    def columns(self) -> list[str]:
        return [f"HistVaR {a:.0%}" for a in self.alpha]


class AVaRMeasureSettings(MeasureSettings):
    type: Literal["Analytical VaR"] = Field("Analytical VaR", repr=False)
    alpha: list[float] = Field([0.1, 0.05])

    @property
    def columns(self) -> list[str]:
        return [f"AnVaR {a:.0%}" for a in self.alpha]


class TimeSeriesVolatilityMeasureSettings(MeasureSettings):
    type: Literal["TimeSeriesVolatility"] = Field("TimeSeriesVolatility", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    ddof: int = Field(1, description="The degrees of freedom to use for the volality")


class CumsumMeasureSettings(MeasureSettings):
    type: Literal["Cumsum"] = Field("Cumsum", repr=False)


class DrawdownMeasureSettings(MeasureSettings):
    type: Literal["Drawdown"] = Field("Drawdown", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")


class MovingAverageMeasureSettings(MeasureSettings):
    type: Literal["MovingAverage"] = Field("MovingAverage", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    var: Literal["r2", "sigma2", "sigma"] = Field(
        "r2", description="The variable to average"
    )

    @property
    def columns(self) -> list[str]:
        return [f"{self.type} {self.var}"]


class FactorMovingAverageMeasureSettings(MeasureSettings):
    type: Literal["FactorMovingAverage"] = Field("FactorMovingAverage", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    var: Literal["abs_t", "p", "p1", "p5"] = Field(
        "p", description="The variable to average"
    )

    @property
    def columns(self) -> list[str]:
        return [f"{self.type} {self.var}"]


class MovingAverageRSquaredMeasureSettings(MeasureSettings):
    type: Literal["MovingAverageRSquared"] = Field("MovingAverageRSquared", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    window: str = Field("6mo", description="The size of the rolling window")


class VolForecastMeasureSettings(MeasureSettings):
    type: Literal["VolForecast"] = Field("VolForecast", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    metric: Literal["qlike", "mse", "mae", "bias"] = Field(
        "qlike", description="The metric to use"
    )
    ddof: int = Field(1, description="The degrees of freedom to use for the stability")

    @property
    def columns(self) -> list[str]:
        return [
            "VolForecast",
            "VolForecastRealized",
            "VolForecastStability",
            "VolForecastLoss",
        ]


class FactorVolForecastMeasureSettings(MeasureSettings):
    type: Literal["FactorVolForecast"] = Field("FactorVolForecast", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    metric: Literal["qlike", "mse", "mae", "bias"] = Field(
        "qlike", description="The metric to use"
    )
    ddof: int = Field(1, description="The degrees of freedom to use for the stability")

    @property
    def columns(self) -> list[str]:
        return [
            "VolForecast",
            "VolForecastRealized",
            "VolForecastStability",
            "VolForecastLoss",
        ]


class FactorCovarianceMeasureSettings(MeasureSettings):
    type: Literal["FactorCovariance"] = Field("FactorCovariance", repr=False)


class IdiosyncraticVolatilityMeasureSettings(MeasureSettings):
    type: Literal["IdiosyncraticVolatility"] = Field(
        "IdiosyncraticVolatility", repr=False
    )


class FactorIdioMeasureSettings(MeasureSettings):
    type: Literal["FactorIdio"] = Field("FactorIdio", repr=False)


class ForecastBacktestMeasureSettings(MeasureSettings):
    type: Literal["ForecastBacktest"] = Field("ForecastBacktest", repr=False)
    window: str = Field("6mo", description="The size of the rolling window")
    ddof: int = Field(1, description="The degrees of freedom to use for the stability")

    @property
    def columns(self) -> list[str]:
        return ["GlobalMinVolPortVolatility", "GlobalMinVolPortPredictedVolatility"]


class BrinsonAttributionMeasureSettings(MeasureSettings):
    type: Literal["BrinsonAttribution"] = Field("BrinsonAttribution", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    multiperiod_aggregation: Literal["none", "optimized"] = Field("none")
    return_aggregation_type: Literal["arithmetic", "geometric"] = Field("geometric")

    @property
    def columns(self) -> list[str]:
        return ["Allocation", "Selection", "Interaction", "Active"]


class FactorAttributionMeasureSettings(MeasureSettings):
    type: Literal["FactorAttribution"] = Field("FactorAttribution", repr=False)
    rescale_bench: bool = Field(
        True,
        description="Rescale the benchmark holdings to sum to the sum of the holdings.",
    )
    normalize_holdings: bool = Field(True, description="Make holdings sum to one.")
    multiperiod_aggregation: Literal["none", "optimized"] = Field("none")
    return_aggregation_type: Literal["arithmetic", "geometric"] = Field("geometric")

    @property
    def columns(self) -> list[str]:
        return ["Portfolio", "Benchmark", "Active"]


class PortfolioStressTestMeasureSettings(MeasureSettings):
    type: Literal["PortfolioStressTest"] = Field("PortfolioStressTest", repr=False)
    shock_size: float = Field(0.1, description="The size of the shock to apply.")

    @property
    def columns(self) -> list[str]:
        return ["Impact"]


class AssetStressTestMeasureSettings(MeasureSettings):
    type: Literal["AssetStressTest"] = Field("AssetStressTest", repr=False)
    shock_size: float = Field(
        1.0, description="The size of the shock to apply in standard deviations."
    )

    @property
    def columns(self) -> list[str]:
        return ["Impact"]


class AssetHoldingsMeasureSettings(MeasureSettings):
    type: Literal["AssetHoldings"] = Field("AssetHoldings", repr=False)

    @property
    def columns(self) -> list[str]:
        return [
            *["Position", "Position Bench", "Position Diff"],
            *["Holdings", "Holdings Bench", "Holdings Diff"],
            *["Weight", "Weight Bench", "Weight Diff"],
            *["Rank", "Rank Bench", "Rank Diff"],
        ]


MeasureSettingsType: TypeAlias = Union[tuple(MeasureSettings.__subclasses__())]  # type: ignore
MeasureSettingsClassType = Type[MeasureSettingsType]

ALL_HOLDINGS_MEASURE_TYPES = [
    HoldingsMeasureSettings,
]
ALL_EXPOSURE_MEASURE_TYPES = [
    ExposureMeasureSettings,
]
ALL_XSR_MEASURE_TYPES = [
    XSigmaRhoMeasureSettings,
]
ALL_TSXSR_MEASURE_TYPES = [
    TimeSeriesXSigmaRhoMeasureSettings,
]
ALL_TS_BETA_MEASURE_TYPES = [
    TimeSeriesBetaMeasureSettings,
]
ALL_VAR_MEASURE_TYPES = [
    HVaRMeasureSettings,
    AVaRMeasureSettings,
]
ALL_FACTOR_TS_MEASURE_TYPES = [
    TimeSeriesVolatilityMeasureSettings,
    CumsumMeasureSettings,
    DrawdownMeasureSettings,
]
ALL_RISK_MODEL_FIT_MEASURE_TYPES = [
    MovingAverageMeasureSettings,
]
ALL_RISK_MODEL_FIT_FACTOR_MEASURE_TYPES = [
    FactorMovingAverageMeasureSettings,
]
ALL_RISK_MODEL_PORTFOLIO_FIT_MEASURE_TYPES = [
    MovingAverageRSquaredMeasureSettings,
]
ALL_FACTOR_FORECAST_LOSS_MEASURE_TYPES = [
    FactorVolForecastMeasureSettings,
]
ALL_FORECAST_LOSS_MEASURE_TYPES = [
    VolForecastMeasureSettings,
]
ALL_FACTOR_COVARIANCE_MEASURE_TYPES = [
    FactorCovarianceMeasureSettings,
]
ALL_IDIOSYNCRATIC_VOLATILITY_MEASURE_TYPES = [
    IdiosyncraticVolatilityMeasureSettings,
]
ALL_FACTOR_IDIO_MEASURE_TYPES = [
    FactorIdioMeasureSettings,
]
ALL_STYLE_CORRELATION_MEASURE_TYPES = [
    PassThroughFactor2DMeasureSettings,
]
ALL_STYLE_INDUSTRY_EXPOSURE_MEASURE_TYPES = [
    PassThroughFactor2DMeasureSettings,
]
ALL_FORECAST_BACKTEST_MEASURE_TYPES = [
    ForecastBacktestMeasureSettings,
]
ALL_BRINSON_ATTRIBUTION_MEASURE_TYPES = [
    BrinsonAttributionMeasureSettings,
]
ALL_FACTOR_ATTRIBUTION_MEASURE_TYPES = [
    FactorAttributionMeasureSettings,
]
ALL_PORTFOLIO_STRESS_TEST_MEASURE_TYPES = [
    PortfolioStressTestMeasureSettings,
]
ALL_ASSET_STRESS_TEST_MEASURE_TYPES = [
    AssetStressTestMeasureSettings,
]
ALL_ASSET_HOLDINGS_MEASURE_TYPES = [
    AssetHoldingsMeasureSettings,
]


def _get_default_measures(
    measure_types: list[MeasureSettingsClassType],
) -> list[MeasureSettingsType]:
    return [m() for m in measure_types]


class ConcreteReportSettings(BaseModel, ABC, frozen=True, extra="forbid"):
    type: str
    lagged_holdings: ClassVar[bool] = False
    measures: list[MeasureSettingsType] = Field(
        description="The measures to include in the report.",
    )


class HoldingsReportSettings(ConcreteReportSettings):
    """Defines settings to build a holdings report."""

    type: Literal["Holdings report"] = Field("Holdings report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_HOLDINGS_MEASURE_TYPES),
        description="The measures to include in the report.",
    )


class ExposureReportSettings(ConcreteReportSettings):
    """Defines settings to build an exposure report."""

    type: Literal["Exposure report"] = Field("Exposure report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_EXPOSURE_MEASURE_TYPES),
        description="The measures to include in the report.",
    )


class XSRReportSettings(ConcreteReportSettings):
    """Defines settings to build an XSR report."""

    type: Literal["XSR report"] = Field("XSR report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_XSR_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class TSXSRReportSettings(ConcreteReportSettings):
    """Defines settings to build a time-series XSR report."""

    type: Literal["TSXSR report"] = Field("TSXSR report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_TSXSR_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class TSBetaReportSettings(ConcreteReportSettings):
    """Defines settings to build a time-series Beta report."""

    type: Literal["TS Beta report"] = Field("TS Beta report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_TS_BETA_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class VaRReportSettings(ConcreteReportSettings):
    """Defines settings to build a VaR report."""

    type: Literal["VaR report"] = Field("VaR report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_VAR_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    lookback: int = Field(1000, description="The lookback period for the estimation.")


class FactorTSReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor time-series report."""

    type: Literal["Factor TS report"] = Field("Factor TS report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_FACTOR_TS_MEASURE_TYPES),  # type: ignore
        description="The measures to include in the report.",
    )


class RiskModelFitReportSettings(ConcreteReportSettings):
    """Defines settings to build a risk model fit report."""

    type: Literal["Risk Model Fit report"] = Field("Risk Model Fit report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_RISK_MODEL_FIT_MEASURE_TYPES),
        description="The measures to include in the report.",
    )


class RiskModelFitFactorReportSettings(ConcreteReportSettings):
    """Defines settings to build a risk model fit report at the factor level."""

    type: Literal["Risk Model Fit Factor report"] = Field(
        "Risk Model Fit Factor report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_RISK_MODEL_FIT_FACTOR_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )


class RiskModelPortfolioFitReportSettings(ConcreteReportSettings):
    """Defines settings to build a risk model fit report at the portfolio level."""

    type: Literal["Risk Model Portfolio Fit report"] = Field(
        "Risk Model Portfolio Fit report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_RISK_MODEL_PORTFOLIO_FIT_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )


class FactorForecastLossReportSettings(ConcreteReportSettings):
    """Defines settings to build a forecast loss report."""

    type: Literal["Factor Forecast Loss report"] = Field(
        "Factor Forecast Loss report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FACTOR_FORECAST_LOSS_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    horizons: list[int] = Field([1, 5, 21], description="The forecast horizons to use.")
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )


class AssetForecastLossReportSettings(ConcreteReportSettings):
    """Defines settings to build a forecast loss report."""

    type: Literal["Forecast Loss report"] = Field("Forecast Loss report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_FORECAST_LOSS_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    horizons: list[int] = Field([1, 5, 21], description="The forecast horizons to use.")
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class FactorCovarianceReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor covariance report."""

    type: Literal["Factor Covariance report"] = Field(
        "Factor Covariance report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FACTOR_COVARIANCE_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )


class IdiosyncraticVolatilityReportSettings(ConcreteReportSettings):
    """Defines settings to build an idiosyncratic volatility report."""

    type: Literal["Idiosyncratic Volatility report"] = Field(
        "Idiosyncratic Volatility report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_IDIOSYNCRATIC_VOLATILITY_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class FactorIdioReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor vcov and idio report."""

    type: Literal["Factor Idio report"] = Field("Factor Idio report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_FACTOR_IDIO_MEASURE_TYPES),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class StyleCorrelationReportSettings(ConcreteReportSettings):
    """Defines settings to build a style correlation report."""

    type: Literal["Style Correlation report"] = Field(
        "Style Correlation report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_STYLE_CORRELATION_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )


class StyleIndustryExposureReportSettings(ConcreteReportSettings):
    """Defines settings to build a style industry exposure report."""

    type: Literal["Style Industry Exposure report"] = Field(
        "Style Industry Exposure report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_STYLE_INDUSTRY_EXPOSURE_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )


class ForecastBacktestReportSettings(ConcreteReportSettings):
    """Defines settings to build a forecast backtest report."""

    type: Literal["Forecast Backtest report"] = Field(
        "Forecast Backtest report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FORECAST_BACKTEST_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    horizons: list[int] = Field([1, 5, 21], description="The forecast horizons to use.")
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class BrinsonAttributionReportSettings(ConcreteReportSettings):
    """Defines settings to build a Brinson attribution report."""

    type: Literal["Brinson Attribution report"] = Field(
        "Brinson Attribution report", repr=False
    )
    lagged_holdings: ClassVar[bool] = True
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_BRINSON_ATTRIBUTION_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )


class FactorAttributionReportSettings(ConcreteReportSettings):
    """Defines settings to build a factor attribution report."""

    type: Literal["Factor Attribution report"] = Field(
        "Factor Attribution report", repr=False
    )
    lagged_holdings: ClassVar[bool] = True
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_FACTOR_ATTRIBUTION_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )


class PortfolioStressTestReportSettings(ConcreteReportSettings):
    """Defines settings to build a portfolio stress test report."""

    type: Literal["Portfolio Stress Test report"] = Field(
        "Portfolio Stress Test report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_PORTFOLIO_STRESS_TEST_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class AssetStressTestReportSettings(ConcreteReportSettings):
    """Defines settings to build an asset stress test report."""

    type: Literal["Asset Stress Test report"] = Field(
        "Asset Stress Test report", repr=False
    )
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(
            ALL_ASSET_STRESS_TEST_MEASURE_TYPES
        ),
        description="The measures to include in the report.",
    )
    halflife_factor_vol: int = Field(
        42, description="The half-life for the factor volatility."
    )
    halflife_factor_adj: int | None = Field(
        None,
        description="The half-life for the factor adjustment. If None, don't adjustment.",
    )
    halflife_factor_cor: int = Field(
        126, description="The half-life for the factor correlation."
    )
    halflife_idio_vol: int = Field(
        42, description="The half-life for the idiosyncratic volatility."
    )
    overlap_factor_vol: int = Field(
        0, description="The overlap for the factor volatility."
    )
    overlap_factor_cor: int = Field(
        0, description="The overlap for the factor correlation."
    )
    overlap_idio_vol: int = Field(
        0, description="The overlap for the idiosyncratic volatility."
    )
    shrink_idio_vol: float = Field(
        0.0, description="The shrinkage factor for the idiosyncratic volatility."
    )


class AssetHoldingsReportSettings(ConcreteReportSettings):
    """Defines settings to build an asset holdings report."""

    type: Literal["Asset Holdings report"] = Field("Asset Holdings report", repr=False)
    measures: list[MeasureSettingsType] = Field(
        default_factory=lambda: _get_default_measures(ALL_ASSET_HOLDINGS_MEASURE_TYPES),
        description="The measures to include in the report.",
    )


ConcreteReportSettingsType: TypeAlias = Union[tuple(ConcreteReportSettings.__subclasses__())]  # type: ignore


class ReportSettings(Settings):
    """
    Defines settings to build a report.
    """

    report: ConcreteReportSettingsType
    risk_model: Annotated[
        str | int | FactorRiskModelSettings,
        Field(
            description="The risk model to use for the report.",
        ),
        SettingsTypeMetaData[str | int | FactorRiskModelSettings](
            references=FactorRiskModelSettings
        ),
    ]


class ConcreteReportSettingsMenu(SettingsMenu, ABC):

    measure_types: ClassVar[list[MeasureSettingsClassType]] = []

    def validate_settings(self, settings: ReportSettings) -> None:
        # TODO implement
        pass

    def describe(self, settings: ReportSettings | None = None) -> str:
        # TODO importment
        return ""


class HoldingsReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_HOLDINGS_MEASURE_TYPES


class ExposureReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_EXPOSURE_MEASURE_TYPES


class XSRReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_XSR_MEASURE_TYPES


class TSXSRReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_TSXSR_MEASURE_TYPES


class TSBetaReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_TS_BETA_MEASURE_TYPES


class VaRReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = ALL_VAR_MEASURE_TYPES


class FactorTSReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_TS_MEASURE_TYPES  # type: ignore
    )


class RiskModelFitReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_RISK_MODEL_FIT_MEASURE_TYPES
    )


class RiskModelFitFactorReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_RISK_MODEL_FIT_FACTOR_MEASURE_TYPES
    )


class RiskModelPortfolioFitReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_RISK_MODEL_FIT_MEASURE_TYPES
    )


class FactorForecastLossReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FORECAST_LOSS_MEASURE_TYPES  # type: ignore
    )


class AssetForecastLossReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FORECAST_LOSS_MEASURE_TYPES  # type: ignore
    )


class FactorCovarianceReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_COVARIANCE_MEASURE_TYPES
    )


class IdiosyncraticVolatilityReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_IDIOSYNCRATIC_VOLATILITY_MEASURE_TYPES
    )


class FactorIdioReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_IDIO_MEASURE_TYPES
    )


class StyleCorrelationReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_STYLE_CORRELATION_MEASURE_TYPES
    )


class StyleIndustryExposureReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_STYLE_INDUSTRY_EXPOSURE_MEASURE_TYPES
    )


class ForecastBacktestReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FORECAST_BACKTEST_MEASURE_TYPES
    )


class BrinsonAttributionReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_BRINSON_ATTRIBUTION_MEASURE_TYPES
    )


class FactorAttributionReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_FACTOR_ATTRIBUTION_MEASURE_TYPES
    )


class PortfolioStressTestReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_PORTFOLIO_STRESS_TEST_MEASURE_TYPES
    )


class AssetStressTestReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_ASSET_STRESS_TEST_MEASURE_TYPES
    )


class AssetHoldingsReportSettingsMenu(ConcreteReportSettingsMenu):
    measure_types: ClassVar[list[MeasureSettingsClassType]] = (
        ALL_ASSET_HOLDINGS_MEASURE_TYPES
    )


class ReportSettingsMenu(SettingsMenu[ReportSettings]):
    """
    Defines available report settings to build a report.
    """

    hld: ConcreteReportSettingsMenu = HoldingsReportSettingsMenu()
    exp: ConcreteReportSettingsMenu = ExposureReportSettingsMenu()
    xsr: ConcreteReportSettingsMenu = XSRReportSettingsMenu()
    tsx: ConcreteReportSettingsMenu = TSXSRReportSettingsMenu()
    bet: ConcreteReportSettingsMenu = TSBetaReportSettingsMenu()
    var: ConcreteReportSettingsMenu = VaRReportSettingsMenu()
    fts: ConcreteReportSettingsMenu = FactorTSReportSettingsMenu()
    rmf: ConcreteReportSettingsMenu = RiskModelFitReportSettingsMenu()
    rms: ConcreteReportSettingsMenu = RiskModelFitFactorReportSettingsMenu()
    rmp: ConcreteReportSettingsMenu = RiskModelPortfolioFitReportSettingsMenu()
    ffl: ConcreteReportSettingsMenu = FactorForecastLossReportSettingsMenu()
    afl: ConcreteReportSettingsMenu = AssetForecastLossReportSettingsMenu()
    fcr: ConcreteReportSettingsMenu = FactorCovarianceReportSettingsMenu()
    ivr: ConcreteReportSettingsMenu = IdiosyncraticVolatilityReportSettingsMenu()
    fni: ConcreteReportSettingsMenu = FactorIdioReportSettingsMenu()
    scr: ConcreteReportSettingsMenu = StyleCorrelationReportSettingsMenu()
    six: ConcreteReportSettingsMenu = StyleIndustryExposureReportSettingsMenu()
    fbt: ConcreteReportSettingsMenu = ForecastBacktestReportSettingsMenu()
    bat: ConcreteReportSettingsMenu = BrinsonAttributionReportSettingsMenu()
    fat: ConcreteReportSettingsMenu = FactorAttributionReportSettingsMenu()
    pst: ConcreteReportSettingsMenu = PortfolioStressTestReportSettingsMenu()
    ast: ConcreteReportSettingsMenu = AssetStressTestReportSettingsMenu()
    ahd: ConcreteReportSettingsMenu = AssetHoldingsReportSettingsMenu()

    def describe(self, settings: ReportSettings | None = None) -> str:
        # TODO
        if settings is not None:
            return os.linesep.join([f"Risk Model: {settings.risk_model}"])
        else:
            return "NA"

    def validate_settings(self, settings: ReportSettings) -> None:
        # TODO
        pass



import abc

from bayesline.api._src.equity.modelconstruction_settings import (
    ModelConstructionSettings,
    ModelConstructionSettingsMenu,
)
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi


class ModelConstructionEngineApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ModelConstructionSettings:
        """
        Returns
        -------
        The modelconstruction settings.
        """
        ...


class AsyncModelConstructionEngineApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ModelConstructionSettings:
        """
        Returns
        -------
        The modelconstruction settings.
        """
        ...


class BayeslineModelConstructionApi(
    RegistryBasedApi[
        ModelConstructionSettings,
        ModelConstructionSettingsMenu,
        ModelConstructionEngineApi,
    ],
): ...


class AsyncBayeslineModelConstructionApi(
    AsyncRegistryBasedApi[
        ModelConstructionSettings,
        ModelConstructionSettingsMenu,
        AsyncModelConstructionEngineApi,
    ],
): ...



import datetime
import json
import os
from collections.abc import Mapping
from typing import Literal

from pydantic import BaseModel, Field, field_validator, model_validator

from bayesline.api._src.equity import settings as settings_tools
from bayesline.api._src.registry import Settings, SettingsMenu
from bayesline.api._src.types import IdType

Hierarchy = settings_tools.Hierarchy


class IndustrySettings(BaseModel, frozen=True, extra="forbid"):
    """
    Specifies include and exclude filters for industries.

    Assets are includes if they are part of at least one include
    and not part of any exclude.

    By default all industries for the given hierarchy are included.
    """

    hierarchy: str = Field(
        default="trbc",
        min_length=1,
        description="The industry hierarchy to use.",
        examples=["trbc"],
    )

    include: list[str] | Literal["All"] = Field(
        default="All",
        description=(
            "Valid industry codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the highest "
            "level is used."
        ),
        examples=[["3571"], "All", ["Materials", "1010"]],
    )

    exclude: list[str] = Field(
        default_factory=list,
        description=(
            "Valid industry codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the lowest "
            "level is used."
        ),
        examples=[["3571"], ["Materials", "1010"]],
    )


class RegionSettings(BaseModel, frozen=True, extra="forbid"):
    """
    Specifies include and exclude filters for countries and regions.

    Assets are includes if they are part of at least one include
    and not part of any exclude.

    By default all countries for the given hierarchy are included.
    """

    hierarchy: str = Field(
        default="continent",
        min_length=1,
        description="The region hierarchy to use.",
        examples=["continent"],
    )

    include: list[str] | Literal["All"] = Field(
        default="All",
        description=(
            "Valid country/region codes or labels at any level. If labels are used "
            "which may be duplicated, then the code with the highest level is used."
        ),
        examples=[["Europe", "CAN"], "All"],
    )

    exclude: list[str] = Field(
        default_factory=list,
        description=(
            "Valid country/region codes or labels at any level. If labels are used "
            "which may be duplicated, then the code with the lowest level is used."
        ),
        examples=[["DEU"]],
    )


class MCapSettings(BaseModel, frozen=True, extra="forbid"):
    """
    Specifies the lower and upper bound for the market cap filter.

    By default the bounds are infinite.
    """

    lower: float = Field(
        default=0,
        ge=0,
        description="Lower bound of the cap filter in USD.",
        examples=[1e10],
    )

    upper: float = Field(
        default=1e99,
        gt=0,
        description="Lower bound of the cap filter in USD.",
        examples=[1e12],
    )

    gdp_deflator_asof: datetime.date | None = Field(
        default=None,
        description="""
        The as of date to adjust the market cap bounds for GDP through time.
        If no date is specified then the market cap bounds are static through time.
        """,
    )

    @model_validator(mode="after")
    def check_upper_gt_lower(self) -> "MCapSettings":
        if (lower := self.lower) >= (upper := self.upper):
            raise ValueError(
                f"upper bound {upper} must be greater than lower bound {lower}",
            )
        else:
            return self


class UniverseSettings(Settings):
    """
    Defines an asset universe as a set of regional, industry and market cap filters.
    """

    @classmethod
    def default(cls: type["UniverseSettings"]) -> "UniverseSettings":
        """
        Creates default universe settings with no filters and
        cusip as the default id type.
        """
        return cls()

    id_type: IdType = Field(
        default="bayesid",
        description="The default id type to use for the universe.",
        examples=["cusip9", "bayesid"],
    )

    industry: IndustrySettings = Field(
        default_factory=IndustrySettings,
        description="""
        Filters that determine which industries to include and exclude in the universe.
        """,
    )

    region: RegionSettings = Field(
        default_factory=RegionSettings,
        description="""
        Filters that determine which countries/continents to include and exclude in the
        universe.
        """,
    )

    mcap: MCapSettings = Field(
        default_factory=MCapSettings,
        description="""
        Filters that determine which market caps to include and exclude in the universe.
        """,
    )


class UniverseSettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """
    Contains the available settings that can be used for the universe settings.
    """

    id_types: list[IdType] = Field(
        description="""
        A list of all the id types that are supported for the universe.
        """,
    )

    industry: Mapping[str, Mapping[str, Hierarchy]] = Field(
        description="""
        A dictionary where the key is the name of the industry hierarchy (e.g. 'GICS')
        and the value is a N-level nested dictionary structure of the industry hierarchy
        codes.
        """,
    )

    region: Mapping[str, Mapping[str, Hierarchy]] = Field(
        description="""
        A dictionary where the key is the name of the region hierarchy (e.g.
        'CONTINENT') and the value is a N-level nested dictionary structure of the
        region hierarchy codes.
        """,
    )

    industry_labels: Mapping[str, Mapping[str, str]] = Field(
        description="""
        A dictionary where the key is the name of the industry hierarchy and
        the value is a mapping from unique industry code to a human readable name.
        """,
    )

    region_labels: Mapping[str, Mapping[str, str]] = Field(
        description="""
        A dictionary where the key is the name of the region hierarchy and
        the value is a mapping from unique region code to a human readable name.
        """,
    )

    def describe(self, settings: UniverseSettings | None = None) -> str:
        industry = self.industry
        region = self.region

        if settings is not None:
            self.validate_settings(settings)

            industry_hierarchy_name = settings.industry.hierarchy
            effective_industries = set(self.effective_industries(settings.industry))
            effective_industry_hierarchy = settings_tools.map_hierarchy(
                settings_tools.filter_leaves(
                    self.industry[industry_hierarchy_name],
                    lambda k: k in effective_industries,
                ),
                self.industry_labels[industry_hierarchy_name].__getitem__,
            )

            region_hierarchy_name = settings.region.hierarchy
            effective_regions = self.effective_regions(settings.region)
            effective_region_hierarchy = settings_tools.map_hierarchy(
                settings_tools.filter_leaves(
                    self.region[region_hierarchy_name],
                    lambda k: k in effective_regions,
                ),
                self.region_labels[region_hierarchy_name].__getitem__,
            )

            description = [
                f"Default ID Type: {settings.id_type!r}",
                f"Industry Hierarchy ({industry_hierarchy_name}):",
                json.dumps(effective_industry_hierarchy, indent=2),
                f"Region Hierarchy ({region_hierarchy_name}):",
                json.dumps(effective_region_hierarchy, indent=2),
                "Market Cap:",
                settings.mcap.model_dump_json(indent=2),
            ]
        else:
            industry_strs = []
            region_strs = []
            for hierarchy in industry:
                industry_str = json.dumps(
                    {
                        hierarchy: settings_tools.map_hierarchy(
                            industry[hierarchy],
                            self.industry_labels[hierarchy].__getitem__,
                        ),
                    },
                    indent=2,
                )
                industry_strs.append(industry_str)

            for hierarchy in region:
                region_str = json.dumps(
                    {
                        hierarchy: settings_tools.map_hierarchy(
                            region[hierarchy],
                            self.region_labels[hierarchy].__getitem__,
                        ),
                    },
                    indent=2,
                )
                region_strs.append(region_str)

            id_types = ", ".join(self.id_types)
            description = [
                f"ID Types: {id_types}",
                "Industry Hierarchies:",
                os.linesep.join(industry_strs),
                "Region Hierarchies:",
                os.linesep.join(region_strs),
            ]

        return os.linesep.join(description)

    @field_validator("industry", "region")
    @classmethod
    def check_unique_hierarchy(
        cls: type["UniverseSettingsMenu"],
        v: Mapping[str, Mapping[str, Hierarchy]],
    ) -> Mapping[str, Mapping[str, Hierarchy]]:
        return settings_tools.check_unique_hierarchy(v)

    @field_validator("industry", "region")
    @classmethod
    def check_nonempty_hierarchy(
        cls: type["UniverseSettingsMenu"],
        v: Mapping[str, Mapping[str, Hierarchy]],
    ) -> Mapping[str, Mapping[str, Hierarchy]]:
        return settings_tools.check_nonempty_hierarchy(v)

    @model_validator(mode="after")
    def check_all_codes_have_labels(self) -> "UniverseSettingsMenu":
        industry_errors = settings_tools.check_all_codes_have_labels(
            self.industry,
            self.industry_labels,
        )
        region_errors = settings_tools.check_all_codes_have_labels(
            self.region,
            self.region_labels,
        )

        errors = industry_errors + region_errors
        if errors:
            raise ValueError(os.linesep.join(errors))
        else:
            return self

    def effective_industries(
        self,
        settings: IndustrySettings,
        labels: bool = False,
    ) -> list[str]:
        """
        Parameters
        ----------
        settings: IndustrySettings
                  the industry settings to get the effective industries for.

        labels: bool
                whether to return the labels or the codes.

        Returns
        -------
        The effective leaf level industries for the given settings after the filters
        were applied.
        """
        self.validate_industry(settings)
        hierarchy = self.industry[settings.hierarchy]
        includes = settings.include
        excludes = settings.exclude
        industry_labels = self.industry_labels[settings.hierarchy]
        effective_codes = self.get_effective_includes(
            hierarchy,
            includes,
            excludes,
            industry_labels,
        )
        if labels:
            return [industry_labels[code] for code in effective_codes]
        else:
            return effective_codes

    def effective_regions(
        self,
        settings: RegionSettings,
        labels: bool = False,
    ) -> list[str]:
        """
        Parameters
        ----------
        settings: RegionSettings
                  the region settings to get the effective regions for.

        labels: bool
                whether to return the labels or the codes.

        Returns
        -------
        The effective leaf level regions for the given settings after the filters were
        applied.
        """
        self.validate_region(settings)
        hierarchy = self.region[settings.hierarchy]
        includes = settings.include
        excludes = settings.exclude
        region_labels = self.region_labels[settings.hierarchy]
        effective_codes = self.get_effective_includes(
            hierarchy,
            includes,
            excludes,
            region_labels,
        )
        if labels:
            return [region_labels[code] for code in effective_codes]
        else:
            return effective_codes

    def validate_settings(self, settings: UniverseSettings) -> None:
        """
        Validates the given universe settings against the available settings.

        Will raise an `ValueError` if settings are invalid.

        Parameters
        ----------
        settings: UniverseSettings
                  the universe settings to validate against.
        """
        if settings.id_type not in self.id_types:
            raise ValueError(
                f"""
                Id type {settings.id_type} does not exist.
                Only {', '.join(self.id_types)} exist.
                """,
            )
        self.validate_industry(settings.industry)
        self.validate_region(settings.region)

    def validate_industry(self, settings: IndustrySettings) -> None:
        """
        Validates the given industry settings against the available settings.

        Will raise an `ValueError` if settings are invalid.

        Parameters
        ----------
        settings: IndustrySettings
                  the industry settings to validate against.
        """
        if settings is None:
            raise ValueError("settings cannot be None")

        available_settings = self.industry
        settings_tools.validate_hierarchy_schema(
            available_settings.keys(),
            settings.hierarchy,
        )
        self._validate_hierarchy(
            available_settings[settings.hierarchy],
            settings.include,
            settings.exclude,
            self.industry_labels[settings.hierarchy],
        )

    def validate_region(self, settings: RegionSettings) -> None:
        """
        Validates the given region settings against the available settings.

        Will raise an `ValueError` if settings are invalid.

        Parameters
        ----------
        settings: RegionSettings
                  the region settings to validate against.
        """
        if settings is None:
            raise ValueError("settings cannot be None")

        available_settings = self.region
        settings_tools.validate_hierarchy_schema(
            available_settings.keys(),
            settings.hierarchy,
        )
        self._validate_hierarchy(
            available_settings[settings.hierarchy],
            settings.include,
            settings.exclude,
            self.region_labels[settings.hierarchy],
        )

    @staticmethod
    def _validate_hierarchy(
        hierarchy: Mapping[str, Hierarchy],
        includes: list[str] | Literal["All"],
        excludes: list[str],
        labels: Mapping[str, str],
    ) -> None:
        flat_hierarchy = settings_tools.flatten(hierarchy)
        label_values = set(labels.values())

        unknown = []
        if includes != "All":
            unknown.extend(
                [
                    include
                    for include in includes
                    if include not in flat_hierarchy and include not in label_values
                ],
            )

        unknown.extend(
            [
                exclude
                for exclude in excludes
                if exclude not in flat_hierarchy and exclude not in label_values
            ],
        )

        if len(unknown) > 0:
            raise ValueError(f"there are unknown items: {', '.join(unknown)}")

        effective_includes = UniverseSettingsMenu.get_effective_includes(
            hierarchy,
            includes,
            excludes,
            labels,
        )
        if len(effective_includes) == 0:
            raise ValueError(
                "the includes and exclude statements lead to an empty set.",
            )

    @staticmethod
    def get_effective_includes(  # noqa: C901
        hierarchy: Mapping[str, Hierarchy],
        includes: list[str] | Literal["All"],
        excludes: list[str],
        labels: Mapping[str, str],
    ) -> list[str]:
        leaf_includes = []
        if isinstance(includes, str) and includes == "All":
            leaf_includes = settings_tools.flatten(hierarchy, only_leaves=True)
        else:
            for include in includes:
                v = settings_tools.find_in_hierarchy(hierarchy, include)
                if v is None:
                    # must be a label then
                    codes = [k for k, v in labels.items() if v == include]
                    if not codes:
                        raise AssertionError(
                            f"{v} is unknown which at this stage should've been verified"
                        )

                    # find the highest level in the hierarchy for possible codes
                    depths = {
                        settings_tools.find_in_hierarchy(
                            hierarchy,
                            code,
                            return_depth=True,
                        ): code
                        for code in codes
                    }
                    if len(depths) != len(codes):
                        raise AssertionError(
                            f"Include {include}: {codes} - {depths}"
                            "illegal, all codes must be represented"
                        )
                    v = settings_tools.find_in_hierarchy(
                        hierarchy,
                        depths[min(depths.keys())],  # type: ignore
                    )

                if isinstance(v, list):
                    leaf_includes.extend(v)
                else:
                    if not isinstance(v, Mapping):
                        raise AssertionError()
                    leaf_includes.extend(
                        settings_tools.flatten(v, only_leaves=True),
                    )

        leaf_excludes = []
        for exclude in excludes:
            v = settings_tools.find_in_hierarchy(hierarchy, exclude)
            if v is None:
                # must be a label then
                codes = [k for k, v in labels.items() if v == exclude]
                if not codes:
                    raise AssertionError(
                        f"{v} is unknown which at this stage should've been verified"
                    )

                # find the lowest level in the hierarchy for possible codes
                depths = {
                    settings_tools.find_in_hierarchy(
                        hierarchy,
                        code,
                        return_depth=True,
                    ): code
                    for code in codes
                }
                v = settings_tools.find_in_hierarchy(
                    hierarchy,
                    depths[max(depths.keys())],  # type: ignore
                )

            if isinstance(v, list):
                leaf_excludes.extend(v)
            else:
                if not isinstance(v, Mapping):
                    raise AssertionError()
                leaf_excludes.extend(settings_tools.flatten(v, only_leaves=True))

        return list(set(leaf_includes) - set(leaf_excludes))



import abc
import datetime as dt
from typing import Literal

import polars as pl

from bayesline.api._src.equity.riskmodels_settings import (
    FactorRiskModelSettings,
    FactorRiskModelSettingsMenu,
)
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi
from bayesline.api._src.types import DateLike, IdType

FactorType = Literal["Market", "Style", "Industry", "Region"]


class FactorRiskModelApi(abc.ABC):

    @abc.abstractmethod
    def dates(self) -> list[dt.date]:
        """
        Returns
        -------
        All dates covered by this risk model.
        """
        pass

    @abc.abstractmethod
    def factors(self, *which: FactorType) -> list[str]:
        """
        Parameters
        ----------
        which: FactorType
            The factor types to return, e.g. `Market`, `Style`, `Industry`, `Region`.
            By default returns all factors.

        Returns
        -------
        list of all factors for the given factor types.
        """
        ...

    @abc.abstractmethod
    def exposures(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the risk model exposures for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range with the first two column as the date and
            asset id. The remaining columns are the individual styles.
        """
        ...

    @abc.abstractmethod
    def universe(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the risk model universe for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the first column is the date and the
            remaining columns are the asset ids. The values are the universe inclusion.
        """
        ...

    @abc.abstractmethod
    def estimation_universe(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the risk model estimation universe for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the first column is the date and the
            remaining columns are the asset ids. The values are the estimation universe
            inclusion.
        """
        ...

    @abc.abstractmethod
    def market_caps(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the market caps for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the index is the date
            and the columns are the asset id. The values are the asset market caps.
        """
        ...

    @abc.abstractmethod
    def future_asset_returns(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the asset returns for this risk model on the next day.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the index is the date
            and the columns are the asset id. The values are the asset returns.
        """
        ...

    @abc.abstractmethod
    def market_stats(
        self,
        estimation_universe: bool = False,
        industries: bool = False,
        regions: bool = False,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        estimation_universe: bool, optional
            If True, returns the market stats for the estimation universe.
        industries: bool, optional
            If True, groups the market by industries.
        regions: bool, optional
            If True, groups the market by regions.

        Returns
        -------
        pl.DataFrame
            Descriptive daily stats for this risk model.
        """
        ...

    @abc.abstractmethod
    def fret(
        self,
        *,
        freq: str | None = None,
        cumulative: bool = False,
        start: DateLike | None = None,
        end: DateLike | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        freq: str, optional
            The frequency of the return aggregation, e.g. `D` for daily.
            Defaults to daily (i.e. unaggregated)
        cumulative: bool, optional
            If True, returns the cumulative returns.
        start: DateLike, optional
        end: DateLike, optional

        Returns
        -------
        pl.DataFrame
            The factor returns for the given date range.
        """
        ...

    @abc.abstractmethod
    def t_stats(self) -> pl.DataFrame: ...

    @abc.abstractmethod
    def p_values(self) -> pl.DataFrame: ...

    @abc.abstractmethod
    def r2(self) -> pl.DataFrame: ...

    @abc.abstractmethod
    def sigma2(self) -> pl.DataFrame: ...

    @abc.abstractmethod
    def style_correlation(
        self, start: DateLike | None = None, end: DateLike | None = None
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        start: DateLike, optional
        end: DateLike, optional

        Returns
        -------
        pl.DataFrame
            The style correlation matrix for the given date range.
        """
        ...

    @abc.abstractmethod
    def industry_exposures(
        self, start: DateLike | None = None, end: DateLike | None = None
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        start: DateLike, optional
        end: DateLike, optional

        Returns
        -------
        pl.DataFrame
            The average style exposures grouped by industry.
        """
        ...


class AsyncFactorRiskModelApi(abc.ABC):

    @abc.abstractmethod
    async def dates(self) -> list[dt.date]:
        """
        Returns
        -------
        All dates covered by this risk model.
        """
        pass

    @abc.abstractmethod
    async def factors(self, *which: FactorType) -> list[str]:
        """
        Parameters
        ----------
        which: FactorType
            The factor types to return, e.g. `Market`, `Style`, `Industry`, `Region`.
            By default returns all factors.

        Returns
        -------
        list of all factors for the given factor types.
        """
        ...

    @abc.abstractmethod
    async def exposures(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the risk model exposures for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the universe to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range with the first two column as the date and
            asset id. The remaining columns are the individual styles.
        """
        ...

    @abc.abstractmethod
    async def universe(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the risk model universe for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the first column is the date and the
            remaining columns are the asset ids. The values are the universe inclusion.
        """
        ...

    @abc.abstractmethod
    async def estimation_universe(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the risk model estimation universe for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the first column is the date and the
            remaining columns are the asset ids. The values are the estimation universe
            inclusion.
        """
        ...

    @abc.abstractmethod
    async def market_caps(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the market caps for this risk model.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the index is the date
            and the columns are the asset id. The values are the asset market caps.
        """
        ...

    @abc.abstractmethod
    async def future_asset_returns(
        self,
        *,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the asset returns for this risk model on the next day.

        Parameters
        ----------
        start: DateLike, optional
            The start date of the data to return, inclusive.
        end: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range where the index is the date
            and the columns are the asset id. The values are the asset returns.
        """
        ...

    @abc.abstractmethod
    async def market_stats(
        self,
        estimation_universe: bool = False,
        industries: bool = False,
        regions: bool = False,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        estimation_universe: bool, optional
            If True, returns the market stats for the estimation universe.
        industries: bool, optional
            If True, groups the market by industries.
        regions: bool, optional
            If True, groups the market by regions.

        Returns
        -------
        pl.DataFrame
            Descriptive daily stats for this risk model.
        """
        ...

    @abc.abstractmethod
    async def fret(
        self,
        *,
        freq: str | None = None,
        cumulative: bool = False,
        start: DateLike | None = None,
        end: DateLike | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        freq: str, optional
            The frequency of the return aggregation, e.g. `D` for daily.
            Defaults to daily (i.e. unaggregated)
        cumulative: bool, optional
            If True, returns the cumulative returns.
        start: DateLike, optional
        end: DateLike, optional

        Returns
        -------
        pl.DataFrame
            The factor returns for the given date range.
        """
        ...

    @abc.abstractmethod
    async def t_stats(self) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def p_values(self) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def r2(self) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def sigma2(self) -> pl.DataFrame: ...


class FactorRiskEngineApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> FactorRiskModelSettings:
        """
        Returns
        -------
        The settings used to create these risk model.
        """
        ...

    @abc.abstractmethod
    def get(self) -> FactorRiskModelApi:
        """

        Returns
        -------
        A built `FactorRiskModelApi` instance for given settings.
        """


class AsyncFactorRiskEngineApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> FactorRiskModelSettings:
        """
        Returns
        -------
        The settings used to create these risk model.
        """
        ...

    @abc.abstractmethod
    async def get(self) -> AsyncFactorRiskModelApi:
        """

        Returns
        -------
        A built `FactorRiskModelApi` instance for given settings.
        """


class BayeslineFactorRiskModelsApi(
    RegistryBasedApi[
        FactorRiskModelSettings, FactorRiskModelSettingsMenu, FactorRiskEngineApi
    ]
): ...


class AsyncBayeslineFactorRiskModelsApi(
    AsyncRegistryBasedApi[
        FactorRiskModelSettings, FactorRiskModelSettingsMenu, AsyncFactorRiskEngineApi
    ]
): ...



import json
import os
from collections import Counter
from collections.abc import Mapping

from pydantic import BaseModel, Field, field_validator, model_validator

from bayesline.api._src.equity import settings as settings_tools
from bayesline.api._src.equity.universe_settings import (
    IndustrySettings,
    RegionSettings,
    UniverseSettings,
    UniverseSettingsMenu,
)
from bayesline.api._src.registry import Settings, SettingsMenu

Hierarchy = settings_tools.Hierarchy

# flake8: noqa: E800


class HierarchyDescription(BaseModel):

    hierarchy: str = Field(
        min_length=1,
        description="""
        The name of the hierarchy to use, e.g. 'trbc' or 'continent'.
        If it is not given then the default hierarchy will be used.
        """,
        examples=["trbc", "continent"],
    )


class HierarchyLevel(HierarchyDescription):
    """
    The hierarchy level description turns every name at
    the configured level into a separate factor.

    E.g. for industries specifying level `1` would
    create top level sector factors.
    """

    level: int = Field(
        description="""The level of the hierarchy to use, e.g. 1
        to use all level 1 names (i.e. sectors for industries or
        continents for regions) or 2 to use all level 2
        names (i.e. sub-sectors for industries and
        countries for regions).
        """,
        default=1,
        examples=[1, 2],
        ge=1,
    )


class HierarchyGroups(HierarchyDescription):
    """
    The hierarchy group description allows for a nested definition
    of groupings.
    The top level groupings will turn into factors, whereas any nested
    groupings will be retained for other uses (e.g. risk decomposition).
    """

    groupings: Mapping[str, Hierarchy] = Field(
        description="""
        A nested structure of groupings where the keys are the group names
        and the leaf level is a list of hierarchy codes or labels (at any level)
        to include for this group.
        """,
    )


class ExposureSettings(Settings):
    """
    Defines exposures as hierarchy of selected styles and substyles.
    """

    @classmethod
    def default(cls: type["ExposureSettings"]) -> "ExposureSettings":
        return cls()

    market: bool = Field(
        default=True,
        description="""
        Whether to include the market factor in the model.
        """,
    )

    styles: Mapping[str, list[str]] | None = Field(
        default=None,
        description="""
        A mapping where the keys are style codes or labels and the values are
        lists of included sub-style names or labels.
        By default (None) the entire available style/substyle hierarchy will be used.
        Passing an empty dict will exclude all styles.
        """,
    )

    industries: HierarchyLevel | HierarchyGroups | None = Field(
        description="""
        The definition of how industry factors are being constructed.
        The default is to use the same hiearchy as was used to define the universe
        at level 1 (i.e. coarse grouping).
        None indicates that no industry factors should be included.
        """,
        default_factory=lambda: HierarchyLevel(hierarchy="trbc", level=1),
    )

    regions: HierarchyLevel | HierarchyGroups | None = Field(
        description="""
        The definition of how region factors are being constructed.
        The default is to use the same hiearchy as was used to define the universe
        at level 2 (i.e. granular grouping).
        None indicates that no region factors should be included.
        """,
        default_factory=lambda: HierarchyLevel(hierarchy="continent", level=2),
    )

    other: Mapping[str, str] = Field(
        description="""
        A mapping of other factors to include in the model, to the exposure names or
        labels. For example, to include the risk free rate as an exposure, the mapping
        could be `{"Risk Free Rate": "us_rate_3m"}`, or 
        `{"Risk-Free-Rate": "3 Month US TBill Rate"}`, the key being the exposure name 
        in the output.
        """,
        default_factory=lambda: {},
    )


class ExposureSettingsMenu(SettingsMenu[ExposureSettings], frozen=True, extra="forbid"):
    """
    Contains the available settings that can be used to define exposures.
    """

    market: str = Field(
        description="""
        The unique market factor code to use in the model.
        """,
    )

    styles: Mapping[str, list[str]] = Field(
        description="""
        A mapping where the key is the name of an exposure style codes (e.g. 'LEVERAGE')
        and the value is a list of available sub-style codes (e.g. 'DEBT_TO_ASSETS')
        """,
    )

    industries: Mapping[str, Mapping[str, Hierarchy]] = Field(
        description="""
        A dictionary where the key is the name of the industry hierarchy (e.g. 'TRBC')
        and the value is a N-level nested dictionary structure of the industry hierarchy
        codes.
        """,
    )

    regions: Mapping[str, Mapping[str, Hierarchy]] = Field(
        description="""
        A dictionary where the key is the name of the region hierarchy (e.g.
        'CONTINENT') and the value is a N-level nested dictionary structure of the
        region hierarchy codes.
        """,
    )

    other: set[str] = Field(
        description="""
        A set of other factor codes that can be included in the model.
        """,
    )

    market_labels: Mapping[str, str] = Field(
        description="""
        A mapping from unique market factor code to a human readable name. This is just 
        one pair, but the name is `market_labels` to be consistent.
        """,
        min_length=1,
        max_length=1,
    )

    style_labels: Mapping[str, str] = Field(
        description="""
        A mapping from unique style/substyle code to a human readable name.
        """,
    )

    industry_labels: Mapping[str, Mapping[str, str]] = Field(
        description="""
        A dictionary where the key is the name of the industry hierarchy and
        the value is a mapping from unique industry code to a human readable name.
        """,
    )

    region_labels: Mapping[str, Mapping[str, str]] = Field(
        description="""
        A dictionary where the key is the name of the region hierarchy and
        the value is a mapping from unique region code to a human readable name.
        """,
    )

    other_labels: Mapping[str, str] = Field(
        description="""
        A mapping from unique other factor code to a human readable name.
        """,
    )

    def effective_industry_hierarchy(
        self,
        filter_settings: IndustrySettings,
        settings: HierarchyLevel | HierarchyGroups | None,
    ) -> Mapping[str, Hierarchy]:
        """
        Parameters
        ----------
        filter_settings: IndustrySettings
                    the settings to use to filter the universe.
        settings: HierarchyLevel | HierarchyGroups | None
                  the settings to use to determine the industry factors.

        Returns
        -------
        A dict structure where the keys are the industry factor names
        and the values are the industry codes that are included in that factor.
        """
        if not settings:
            return {}

        effective_includes = UniverseSettingsMenu.get_effective_includes(
            self.industries[settings.hierarchy],
            filter_settings.include,
            filter_settings.exclude,
            self.industry_labels[settings.hierarchy],
        )

        return ExposureSettingsMenu._effective_hierarchy(
            self.industries,
            self.industry_labels,
            settings,
            effective_includes,
        )

    def effective_region_hierarchy(
        self,
        filter_settings: RegionSettings,
        settings: HierarchyLevel | HierarchyGroups | None,
    ) -> Mapping[str, Hierarchy]:
        """
        Parameters
        ----------
        filter_settings: RegionSettings
                the settings to use to filter the universe.
        settings: HierarchyLevel | HierarchyGroups | None
                  the settings to use to determine the region factors.

        Returns
        -------
        A dict structure where the keys are the region factor names
        and the values are the region codes that are included in that factor.
        """
        if not settings:
            return {}

        effective_includes = UniverseSettingsMenu.get_effective_includes(
            self.regions[settings.hierarchy],
            filter_settings.include,
            filter_settings.exclude,
            self.region_labels[settings.hierarchy],
        )

        return ExposureSettingsMenu._effective_hierarchy(
            self.regions,
            self.region_labels,
            settings,
            effective_includes,
        )

    def normalize(
        self,
        universe_settings: UniverseSettings,
        exposure_settings: ExposureSettings,
    ) -> ExposureSettings:
        """
        Normalizes the given exposure settings by converting all
        style and substyle codes/labels to their corresponding codes/labels.

        Parameters
        ----------
        universe_settings: UniverseSettings
                the universe settings to use for normalization.
        exposure_settings: ExposureSettings
                the exposure settings to normalize.

        Returns
        -------
        A new exposure settings object with all style and
        substyle code/labels converted to codes/labels.
        """
        if exposure_settings.industries and not (
            exposure_settings.industries.hierarchy
            == universe_settings.industry.hierarchy
        ):
            raise AssertionError(
                "industry schema for universe filter must be identical to schema for "
                f"industry exposures, but is {universe_settings.industry.hierarchy} vs "
                f"{exposure_settings.industries.hierarchy}"
            )

        if exposure_settings.regions and not (
            exposure_settings.regions.hierarchy == universe_settings.region.hierarchy
        ):
            raise AssertionError(
                "region schema for universe filter must be identical to schema for "
                f"region exposures, but is {universe_settings.region.hierarchy} vs "
                f"{exposure_settings.regions.hierarchy}"
            )

        self.validate_settings(exposure_settings)

        effective_industry_includes = UniverseSettingsMenu.get_effective_includes(
            self.industries[universe_settings.industry.hierarchy],
            universe_settings.industry.include,
            universe_settings.industry.exclude,
            self.industry_labels[universe_settings.industry.hierarchy],
        )
        self.validate_hierarchy(
            settings_tools.filter_leaves(
                self.industries, lambda code: code in effective_industry_includes
            ),
            self.industry_labels,
            exposure_settings.industries,
            enforce_full=True,
        )

        effective_region_includes = UniverseSettingsMenu.get_effective_includes(
            self.regions[universe_settings.region.hierarchy],
            universe_settings.region.include,
            universe_settings.region.exclude,
            self.region_labels[universe_settings.region.hierarchy],
        )
        self.validate_hierarchy(
            settings_tools.filter_leaves(
                self.regions, lambda code: code in effective_region_includes
            ),
            self.region_labels,
            exposure_settings.regions,
            enforce_full=True,
        )

        if exposure_settings.styles is None:
            # choose entire style hierarchy by default
            exposure_settings = exposure_settings.model_copy(
                update={"styles": self.styles}
            )

        normalized_styles = self.normalize_styles(exposure_settings.styles)

        if exposure_settings.industries is None:
            normalized_industries = None
        else:
            normalized_industries = (
                HierarchyGroups(
                    hierarchy=exposure_settings.industries.hierarchy,
                    groupings=self.effective_industry_hierarchy(
                        universe_settings.industry,
                        exposure_settings.industries,
                    ),
                )
                if exposure_settings.industries is not None
                else None
            )

        if exposure_settings.regions is None:
            normalized_regions = None
        else:
            normalized_regions = HierarchyGroups(
                hierarchy=exposure_settings.regions.hierarchy,
                groupings=self.effective_region_hierarchy(
                    universe_settings.region,
                    exposure_settings.regions,
                ),
            )

        lookup = {label: code for code, label in self.other_labels.items()}
        normalized_other = {
            k: lookup.get(v, v) for k, v in exposure_settings.other.items()
        }

        return ExposureSettings(
            market=exposure_settings.market,
            styles=normalized_styles,
            industries=normalized_industries,
            regions=normalized_regions,
            other=normalized_other,
        )

    def normalize_styles(
        self, styles: Mapping[str, list[str]] | None
    ) -> dict[str, list[str]]:
        styles = self.styles if styles is None else styles  # styles={} is valid
        lookup = {label: code for code, label in self.style_labels.items()}

        normalized_styles = {}
        for style, substyles in styles.items():  # type: ignore
            normalized_style = lookup.get(style, style)

            normalized_substyles = []
            for substyle in substyles:
                normalized_substyle = lookup.get(substyle, substyle)
                normalized_substyles.append(normalized_substyle)

            normalized_styles[normalized_style] = normalized_substyles
        return normalized_styles

    def all_substyles(
        self,
        settings: ExposureSettings | None = None,
        *,
        labels: bool = False,
    ) -> list[str]:
        """
        Parameters
        ----------
        settings: ExposureSettings, optional
                  the exposure settings to get all substyles for.
        labels: bool, optional
                whether to return the substyles as labels or codes.

        Returns
        -------
        A sorted flat list of all substyles in this settings menu or all
        configured substyles if a settings object is given.
        """
        result: list[str] = []

        if settings is not None and settings.styles is None:
            # choose entire style hierarchy by default
            settings = settings.model_copy(update={"styles": self.styles})

        if settings is not None:
            self.validate_settings(settings)
            for substyles in settings.styles.values():  # type: ignore
                result.extend(substyles)
        else:
            for substyles in self.styles.values():
                result.extend(substyles)

        if labels:
            result = [
                self.style_labels.get(substyle, substyle) for substyle in set(result)
            ]
        else:
            labels_to_codes = {label: code for code, label in self.style_labels.items()}
            result = [
                labels_to_codes.get(substyle, substyle) for substyle in set(result)
            ]
        return sorted(result)

    @staticmethod
    def _effective_hierarchy(
        hierarchies: Mapping[str, Mapping[str, Hierarchy]],
        labels: Mapping[str, Mapping[str, str]],
        settings: HierarchyLevel | HierarchyGroups,
        effective_includes: list[str],
    ) -> Mapping[str, Hierarchy]:
        if isinstance(settings, HierarchyLevel):
            return ExposureSettingsMenu._effective_hierarchy_from_hierarchy_level(
                hierarchies,
                labels,
                settings,
                effective_includes,
            )
        elif isinstance(settings, HierarchyGroups):
            return ExposureSettingsMenu._effective_hierarchy_from_hierarchy_groups(
                hierarchies,
                labels,
                settings,
                effective_includes,
            )
        else:
            raise NotImplementedError(type(settings))

    @staticmethod
    def _effective_hierarchy_from_hierarchy_level(
        hierarchies: Mapping[str, Mapping[str, Hierarchy]],
        labels: Mapping[str, Mapping[str, str]],
        settings: HierarchyLevel,
        effective_includes: list[str],
    ) -> Mapping[str, Hierarchy]:
        if settings.hierarchy is None:
            raise AssertionError(
                """
                Hierarchy must be given, use `normalize` to
                populate the default hierarchy name.
                """,
            )
        hierarchy = hierarchies[settings.hierarchy]
        hierarchy_labels = labels[settings.hierarchy]

        hierarchy_filtered = settings_tools.filter_leaves(
            hierarchy,
            lambda code: code in effective_includes,
        )

        hierarchy_trimmed = settings_tools.trim_to_depth(
            hierarchy_filtered,
            settings.level,
        )

        groups = settings_tools.flatten({"dummy": hierarchy_trimmed}, only_leaves=True)
        result = {}
        for group in groups:
            sub_hierarchy = settings_tools.find_in_hierarchy(hierarchy, group)
            if sub_hierarchy is None or isinstance(sub_hierarchy, int):
                raise AssertionError(f"could not find {group} in {hierarchy}")

            if isinstance(sub_hierarchy, list):
                result[hierarchy_labels[group]] = [group]
            else:
                result[hierarchy_labels[group]] = settings_tools.flatten(
                    sub_hierarchy[group],
                    only_leaves=True,
                )
        return result

    @staticmethod
    def _effective_hierarchy_from_hierarchy_groups(  # noqa: C901
        hierarchies: Mapping[str, Hierarchy],
        labels: Mapping[str, Mapping[str, str]],
        settings: HierarchyGroups,
        effective_includes: list[str] | None = None,
    ) -> Mapping[str, Hierarchy]:
        if settings.hierarchy is None:
            raise AssertionError(
                """
                Hierarchy must be given, use `normalize` to
                populate the default hierarchy name.
                """,
            )

        hierarchy: Mapping[str, Hierarchy] = {
            settings.hierarchy: hierarchies[settings.hierarchy]
        }
        hierarchy_labels = labels[settings.hierarchy]

        hierarchy_filtered = hierarchy
        if effective_includes is not None:
            hierarchy_filtered = settings_tools.filter_leaves(
                hierarchy,
                lambda code: code in effective_includes,
            )

        # leaf codes in the grouping description might be higher level codes
        # or labels, e.g.
        # {
        #   "Energy": ["Energy"],
        #   "Materials": ["15"],
        #   "Exchanges": ["201010", "201020"],
        #   "Asset Managers": ["201010"],
        #   "Banks": ["201020"],
        # }
        #
        # so first we map all leaf codes/labels to the lowest level codes
        # in the actual hierarchy
        groupings = settings.groupings
        actual_leaves = settings_tools.flatten(hierarchy_filtered, only_leaves=True)

        def _expand_leaves_to_actual_leaves(code_or_label: str) -> list[str]:
            if code_or_label in actual_leaves:
                return [code_or_label]
            else:
                v = settings_tools.find_in_hierarchy(hierarchy_filtered, code_or_label)
                if v is None:
                    # must be a label then
                    codes = [
                        k for k, v in hierarchy_labels.items() if v == code_or_label
                    ]
                    if not codes:
                        raise ValueError(
                            "Unknown hierarchy item for hierarchy "
                            f"{settings.hierarchy}: {code_or_label}"
                        )

                    # find the highest level in the hierarchy for possible codes
                    depths = {
                        settings_tools.find_in_hierarchy(
                            hierarchy_filtered,
                            code,
                            return_depth=True,
                        ): code
                        for code in codes
                    }
                    if len(depths) != len(
                        codes,
                    ):
                        raise NotImplementedError(
                            "{codes} - {depths} illegal, all codes my be represented"
                        )
                    v = settings_tools.find_in_hierarchy(
                        hierarchy_filtered,
                        depths[min(depths.keys())],  # type: ignore
                    )
                if isinstance(v, list):
                    return v
                else:
                    if not isinstance(v, Mapping):
                        raise AssertionError(type(v))
                    return settings_tools.flatten(v, only_leaves=True)

        def _map_leaves(
            h: Mapping[str, Hierarchy] | list[str],
        ) -> Mapping[str, Hierarchy] | list[str]:
            if isinstance(h, list):
                return [
                    e
                    for sublist in [_expand_leaves_to_actual_leaves(e) for e in h]
                    for e in sublist
                ]
            else:
                return {k: _map_leaves(v) for k, v in h.items()}

        effective_groupings = _map_leaves(groupings)
        mapped_grouping_leaves = settings_tools.flatten(
            effective_groupings, only_leaves=True
        )
        settings_tools.check_all_leafcodes_exist(
            {settings.hierarchy: effective_groupings}, set(actual_leaves)
        )
        settings_tools.check_unique_hierarchy(
            {settings.hierarchy: {"__DUMMY__": mapped_grouping_leaves}}
        )

        if not isinstance(effective_groupings, Mapping):
            raise AssertionError(type(effective_groupings))
        return effective_groupings

    def describe(self, settings: ExposureSettings | None = None) -> str:
        if settings is not None:
            self.validate_settings(settings)
            hierarchy = settings.styles or self.styles
            other = [
                self.other_labels.get(other, other) for other in settings.other.values()
            ]
        else:
            hierarchy = self.styles
            other = list(self.other_labels.values())

        style_hierarchy = {}
        for style, substyles in hierarchy.items():
            style_label = self.style_labels.get(style, style)
            style_hierarchy[style_label] = [
                self.style_labels.get(substyle, substyle) for substyle in substyles
            ]

        description = [
            "Style Hierarchy:",
            json.dumps(style_hierarchy, indent=2),
            "",
            "Other Exposures:",
            f"{os.linesep.join([f'  - {o}' for o in sorted(other)])}",
        ]

        return os.linesep.join(description)

    @field_validator("industries", "regions", "styles")
    @classmethod
    def check_unique_hierarchy(
        cls: type["ExposureSettingsMenu"],
        v: Mapping[str, Mapping[str, Hierarchy]],
    ) -> Mapping[str, Mapping[str, Hierarchy]]:
        return settings_tools.check_unique_hierarchy(v)

    @field_validator("industries", "regions")
    @classmethod
    def check_nonempty_hierarchy(
        cls: type["ExposureSettingsMenu"],
        v: Mapping[str, Mapping[str, Hierarchy]],
    ) -> Mapping[str, Mapping[str, Hierarchy]]:
        return settings_tools.check_nonempty_hierarchy(v)

    @field_validator("styles")
    @classmethod
    def check_nonempty_hierarchy_if_provided(
        cls: type["ExposureSettingsMenu"],
        v: Mapping[str, Mapping[str, Hierarchy]],
    ) -> Mapping[str, Mapping[str, Hierarchy]]:
        if len(v) == 0:
            return v
        return settings_tools.check_nonempty_hierarchy(v)

    @model_validator(mode="after")
    def check_all_codes_have_labels(self) -> "ExposureSettingsMenu":
        market_errors = settings_tools.check_all_codes_have_labels(
            {"market": [self.market]},
            {"market": self.market_labels},
        )
        industry_errors = settings_tools.check_all_codes_have_labels(
            self.industries,
            self.industry_labels,
        )
        region_errors = settings_tools.check_all_codes_have_labels(
            self.regions,
            self.region_labels,
        )
        style_errors = settings_tools.check_all_codes_have_labels(
            {"styles": self.styles},
            {"styles": self.style_labels},
        )
        other_errors = settings_tools.check_all_codes_have_labels(
            {"other": list(self.other)},
            {"other": self.other_labels},
        )

        errors = (
            market_errors
            + industry_errors
            + region_errors
            + style_errors
            + other_errors
        )
        if errors:
            raise ValueError(os.linesep.join(errors))
        else:
            return self

    def validate_settings(self, settings: ExposureSettings) -> None:
        """
        Validates the given exposure settings against the available settings.

        Will raise an `ValueError` if settings are invalid.

        Parameters
        ----------
        settings: ExposureSettings
                  the exposure settings to validate against.
        """
        self.validate_styles(settings.styles)
        self.validate_hierarchy(
            self.industries,
            self.industry_labels,
            settings.industries,
            enforce_full=False,
        )
        self.validate_hierarchy(
            self.regions, self.region_labels, settings.regions, enforce_full=False
        )
        self.validate_other(settings.other)

    def validate_styles(self, settings: Mapping[str, list[str]] | None) -> None:
        """
        Validates the given style settings against the available settings.

        Will raise an `ValueError` if settings are invalid.

        Parameters
        ----------
        settings: Mapping[str, list[str]]
                  the style settings to validate against.
        """
        if settings is None:
            # nothing to do, default styles will be used
            return

        available_settings = self.styles
        available_labels = self.style_labels.values()

        error_messages = []
        for style, substyles in settings.items():
            if style not in available_settings and style not in available_labels:
                error_messages.append(f"Style {style} does not exist.")
                continue

            if style in available_labels:
                style_code = next(
                    style_code
                    for style_code, style_label in self.style_labels.items()
                    if style_label == style
                )
            else:
                style_code = style

            available_substyles = available_settings[style_code]

            for substyle in substyles:
                if (
                    substyle not in available_substyles
                    and substyle not in available_labels
                ):
                    error_messages.append(f"Substyle {substyle} does not exist.")

            if len(substyles) > len(set(substyles)):
                dupes = {
                    substyle for substyle in substyles if substyles.count(substyle) > 1
                }
                error_messages.append(
                    f"Substyles {', '.join(dupes)} are duplicated for style {style}.",
                )

        if error_messages:
            raise ValueError(os.linesep.join(error_messages))

    @staticmethod
    def validate_hierarchy(
        hierarchies: Mapping[str, Hierarchy],
        labels: Mapping[str, Mapping[str, str]],
        settings: HierarchyLevel | HierarchyGroups | None,
        enforce_full: bool,
    ) -> None:
        if settings is None:
            return

        settings_tools.validate_hierarchy_schema(hierarchies.keys(), settings.hierarchy)

        if isinstance(settings, HierarchyLevel):
            depth = settings_tools.get_depth(hierarchies[settings.hierarchy])
            if depth < settings.level:
                raise ValueError(
                    f"Illegal level {settings.level}, maximum level is {depth}"
                )
        elif isinstance(settings, HierarchyGroups):
            effective_hierarchy = (
                ExposureSettingsMenu._effective_hierarchy_from_hierarchy_groups(
                    hierarchies,
                    labels,
                    settings,
                )
            )

            existing = set(
                settings_tools.flatten(
                    hierarchies[settings.hierarchy], only_leaves=True
                )
            )

            settings_tools.check_all_leafcodes_exist(effective_hierarchy, existing)

            # the hierarchy is a grouping where the leaves map to the actual industry
            # or region hierarchy and anything above are labels for the grouping
            # hence to check for uniqueness we check anything above the leaves
            # and the leaves themselves
            check_hierarchy = {settings.hierarchy: effective_hierarchy}
            check_hierarchy_no_leaves = settings_tools.prune_leaves(check_hierarchy)
            check_hierarchy_only_leaves = settings_tools.flatten(
                check_hierarchy, only_leaves=True
            )
            settings_tools.check_unique_hierarchy(
                {"__DUMMY__": check_hierarchy_no_leaves}
            )

            dupes = [
                item
                for item, count in Counter(check_hierarchy_only_leaves).items()
                if count > 1
            ]
            if dupes:
                raise ValueError(f"Duplicate hierarchy codes: {', '.join(dupes)}")

            # check all codes are used
            if enforce_full:
                configured_codes = set(
                    settings_tools.flatten(effective_hierarchy, only_leaves=True)
                )
                missing = existing - configured_codes
                if missing:
                    raise ValueError(
                        f"Missing hierarchy codes for {settings.hierarchy}: {','.join(missing)}"
                    )
        else:
            raise ValueError(f"unknown settings type {type(settings)}")

    def validate_other(self, settings: Mapping[str, str]) -> None:
        """
        Validates the given other settings against the available settings.

        Will raise an `ValueError` if settings are invalid.

        Parameters
        ----------
        settings: Mapping[str, str]
            the other settings to validate against.
        """
        available_settings = self.other
        available_labels = self.other_labels.values()

        error_messages = []
        for setting in settings.values():
            if setting not in available_settings and setting not in available_labels:
                error_messages.append(f"Other {setting} does not exist.")

        if error_messages:
            raise ValueError(os.linesep.join(error_messages))



import math
import os
from typing import Literal, Mapping

from pydantic import Field, NonNegativeFloat, field_validator

from bayesline.api._src.registry import Settings, SettingsMenu

WeightingScheme = Literal["SqrtCap", "InvIdioVar"]


class ModelConstructionSettings(Settings):
    """
    Defines settings to build a factor risk model.
    """

    @classmethod
    def default(cls) -> "ModelConstructionSettings":
        return cls(weights="SqrtCap")

    weights: WeightingScheme = Field(
        description="The regression weights used for the factor risk model.",
        default="SqrtCap",
        examples=["SqrtCap", "InvIdioVar"],
    )
    alpha: NonNegativeFloat = Field(
        description="The ridge-shrinkage factor for the factor risk model.",
        default=0.0,
    )
    return_clip_bounds: tuple[float, float] = Field(
        description="The bounds for the return clipping.",
        default=(-0.1, 0.1),
        examples=[(-0.1, 0.1), (-math.inf, math.inf)],
    )
    known_factors: Mapping[str, str] = Field(
        description=(
            "The known factor returns to use for the factor risk model which are added "
            "as independent factors with known (constrained) values. "
            "The keys are the names that should be used in the factor model and the "
            "values are the underlying known factors that should be used. "
        ),
        default_factory=dict,
        examples=[{}, {"Risk Free Rate": "us_rate_3m"}],
    )

    @field_validator("return_clip_bounds")
    @classmethod
    def return_clip_bounds_valid(cls, v: tuple[float, float]) -> tuple[float, float]:
        lb, ub = v
        if ub < lb:
            raise ValueError(f"Lower bound {lb} cannot be bigger than upper bound {ub}")
        return v


class ModelConstructionSettingsMenu(
    SettingsMenu[ModelConstructionSettings], frozen=True, extra="forbid"
):
    """
    Defines available modelconstruction settings to build a factor risk model.
    """

    weights: list[WeightingScheme] = Field(
        description="""
        The available regression weights that can be used for the factor risk model.
        """,
    )

    known_factors: dict[str, str] = Field(
        description="""
        The available known factor returns that can be used for the factor risk model.
        The keys are the factor names and the values a description for this known factor.
        """,
    )

    def describe(self, settings: ModelConstructionSettings | None = None) -> str:
        if settings is not None:
            known_factors = []
            for factor, value in settings.known_factors.items():
                known_factors.append(f"{factor}: {value}")
            out = f"Weights: {settings.weights}"
            if known_factors:
                out += os.linesep
                out += f"Known Factors: {', '.join(known_factors)}"
            return out
        else:
            known_factors = []
            for factor, desc in self.known_factors.items():
                known_factors.append(f"  - {factor}: {desc}")
            out = f"Weights: {', '.join(self.weights)}"
            if known_factors:
                out += os.linesep
                out += f"Known Factors:{os.linesep}{os.linesep.join(known_factors)}"
            return out

    def validate_settings(self, settings: ModelConstructionSettings) -> None:
        if settings.weights not in self.weights:
            raise ValueError(f"Invalid weights: {settings.weights}")

        missing_known_factors = set(settings.known_factors.values()) - set(
            self.known_factors.keys()
        )
        if missing_known_factors:
            raise ValueError(
                f"Invalid known factor returns: {', '.join(missing_known_factors)}"
            )



from __future__ import annotations

import abc
import datetime as dt
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, Generic, Literal, TypeVar

from pydantic import BaseModel, Field, ValidationError


class InvalidSettingsError(ValueError):
    pass


class Settings(BaseModel, frozen=True, extra="forbid"):

    def references(
        self,
    ) -> dict[str, dict[type[Settings], list[str | int]]]:
        """
        All other settings that are referenced by this settings,
        e.g. if a risk model references a unvierse.

        Returns
        -------
        A dict of the field name to a dict of the referenced settings type to the list
        of referenced settings (either by name or id).
        """
        result: dict[str, dict[type[Settings], list[str | int]]] = defaultdict(
            lambda: defaultdict(list)
        )

        for field_name, field_info in self.model_fields.items():
            for meta_data in field_info.metadata:
                if isinstance(meta_data, SettingsTypeMetaData):
                    ref_type = meta_data.references
                    field_value = getattr(self, field_name)
                    if field_value is not None:
                        result[field_name][ref_type].extend(
                            meta_data.extract(field_value)
                        )

        return result


E = TypeVar("E")


@dataclass
class SettingsTypeMetaData(Generic[E]):

    references: type[Settings]
    extractor: Callable[[E], list[str | int]] | None = None

    def extract(self, v: E) -> list[str | int]:
        if self.extractor is None:
            if isinstance(v, str | int):
                return [v]
            elif isinstance(v, self.references):
                return []
            else:
                raise ValueError(
                    f"Cannot extract {v} of type {type(v)}. "
                    "Define an extractor in the pydantic model"
                )
        return self.extractor(v)


class RawSettings(BaseModel):

    model_type: str
    name: str | None
    identifier: int | None
    exists: bool
    raw_json: dict[str, Any]
    references: list[RawSettings]
    extra: dict[str, Any] = Field(default_factory=dict)


T = TypeVar("T")
SettingsType = TypeVar("SettingsType", bound=Settings)
ModelType = str


class AsyncSettingsResolver(abc.ABC):

    @classmethod
    @abc.abstractmethod
    def model_types(cls) -> dict[ModelType, type[Settings]]:
        """
        Returns
        -------
        A dictionary of the model types (str) to their type instance.
        Typically this is e.g. `{UniverseSettings.__name__: UniverseSettings}`.
        """

    @abc.abstractmethod
    async def resolve_settings(self, settings: list[Settings]) -> list[RawSettings]:
        """
        Resolves the given references.

        Parameters
        ----------
        settings: Settings
            The settings to resolve.

        Returns
        -------
        List of raw settings exactly in the same order as the input.
        """

    @abc.abstractmethod
    async def resolve_references(
        self, references: list[tuple[ModelType, str | int]]
    ) -> list[RawSettings]:
        """
        Resolves the given settings references.

        Parameters
        ----------
        references: list[tuple[ModelType, str | int]]
            The settings references to resolve.

        Returns
        -------
        List of raw settings exactly in the same order as the input.
        """

    def to_settings_model(
        self, raw_settings: RawSettings, resolve: bool = False
    ) -> Settings:
        """
        Converts the given raw settings to a settings model.

        Parameters
        ----------
        raw_settings: RawSettings
            The raw settings to convert.
        resolve: bool
            If True, the references are resolved.

        Returns
        -------
        The settings model.
        """
        if resolve:
            # resolve all references recursively
            raise NotImplementedError()
        if not self.is_valid(raw_settings):
            raise InvalidSettingsError(raw_settings.raw_json)
        return self.model_types()[raw_settings.model_type].model_validate(
            raw_settings.raw_json
        )

    def is_valid(self, raw_setting: RawSettings) -> bool:
        """
        Checks if the given raw settings are valid.

        Parameters
        ----------
        raw_setting: RawSettings
            The raw settings to check.

        Returns
        -------
        True if the raw setting is valid, False otherwise.
        """
        valid = raw_setting.exists
        valid &= not self.is_corrupted(raw_setting)
        valid &= self.has_all_references(raw_setting)
        return valid

    def is_corrupted(self, raw_setting: RawSettings) -> bool:
        """
        Checks if the given raw settings are corrupted.

        Parameters
        ----------
        raw_setting: RawSettings
            The raw settings to check.

        Returns
        -------
        True if the raw setting is corrupted, False otherwise.
        """
        if not raw_setting.raw_json:
            return True

        try:
            settings_type = self.model_types()[raw_setting.model_type]
            settings_type.model_validate(raw_setting.raw_json)
            return False
        except ValidationError:
            return True
        except KeyError:
            return True

    def has_all_references(self, raw_settings: RawSettings) -> bool:
        """
        Checks if the given settings have all their references.

        Parameters
        ----------
        settings: Settings
            The raw_settings to check.

        Returns
        -------
        True if the settings have all their references, False otherwise.
        """
        return all(self.is_valid(ref) for ref in raw_settings.references)

    def validation_messages(self, raw_settings: RawSettings) -> list[str]:
        messages = []
        corrupted = self.is_corrupted(raw_settings)
        missing = not raw_settings.exists
        if missing:
            messages.append("The setting does not exist.")
        elif corrupted:
            messages.append("The setting json is invalid.")

        for ref in raw_settings.references:
            ref_messages = [
                m.replace("The setting", f"The reference {ref.name!r}")
                for m in self.validation_messages(ref)
            ]
            messages.extend(ref_messages)
        return messages


class SettingsMetaData(BaseModel, frozen=True, extra="allow"):

    created_on: dt.datetime
    last_updated: dt.datetime


class SettingsMenu(
    abc.ABC,
    BaseModel,
    Generic[SettingsType],
    frozen=True,
    extra="forbid",
):

    @abc.abstractmethod
    def describe(self, settings: SettingsType | None = None) -> str:
        """
        Parameters
        ----------
        settings : SettingsType | None
                   The settings to describe.
                   If None, then the description is not evaluated against any settings.

        Returns
        -------
        A human readable description of the settings menu,
        optionally evaluated against the given settings.
        """

    @abc.abstractmethod
    def validate_settings(self, settings: SettingsType) -> None:
        """
        Validates if the given settings are valid for this settings menu.

        Parameters
        ----------
        settings : SettingsType
                   The settings to validate.

        Raises
        ------
        ValidationError if invalid.
        """


SettingsMenuType = TypeVar("SettingsMenuType", bound=SettingsMenu)

Mode = Literal["All", "Valid", "Invalid"]


class ReadOnlyRegistry(abc.ABC, Generic[T]):

    @abc.abstractmethod
    def ids(self, mode: Mode = "Valid") -> dict[int, str]:
        """
        Parameters
        ----------
        mode : Mode
               The mode to use when retrieving the ids.

        Returns
        -------
        A dictionary of the unique identifiers to the unique names.
        """

    @abc.abstractmethod
    def names(self, mode: Mode = "Valid") -> dict[str, int]:
        """
        Parameters
        ----------
        mode : Mode
               The mode to use when retrieving the ids.

        Returns
        -------
        A dictionary of the unique names to the unique identifiers.
        """

    @abc.abstractmethod
    def get_raw(self, name_or_id: list[str | int]) -> list[RawSettings]:
        """
        Parameters
        ----------
        name_or_id : list[str | int]
                     The unique names or int identifiers of the items to retrieve.

        Returns
        -------
        A list of `RawSettings` in the same order as input.
        """

    @abc.abstractmethod
    def get(self, name: str | int) -> T:
        """
        Parameters
        ----------
        name : str | int
               The unique name or int identifier of the item to retrieve.

        Raises
        ------
        KeyError
            If the item does not exist.
        InvalidSettingsError:
            If the item exists but is invalid.

        Returns
        -------
        The item for the given name.
        """

    @abc.abstractmethod
    def get_metadata(self, name: str | int) -> SettingsMetaData:
        """
        Parameters
        ----------
        name : str | int
               The unique name or int identifier of the item to retrieve.

        Raises
        ------
        KeyError
            If the item does not exist.

        Returns
        -------
        The metadata for the given name.
        """

    def get_all(self) -> dict[str, T]:
        """
        Returns
        -------
        A dictionary of all valid available settings.
        """
        return {name: self.get(name) for name in self.names()}

    def get_all_with_metadata(self) -> dict[str, tuple[T, SettingsMetaData]]:
        """
        Returns
        -------
        A dictionary of all valid available settings with metadata.
        """
        all_settings = self.get_all()
        all_metadata = self.get_all_metadata()

        return {
            name: (all_settings[name], all_metadata[name])
            for name in all_settings.keys()
        }

    def get_all_metadata(self) -> dict[str, SettingsMetaData]:
        """
        Returns
        -------
        A dictionary of all available settings metadata (valid or invalid).
        """
        return {name: self.get_metadata(name) for name in self.names()}


class Registry(Generic[T], ReadOnlyRegistry[T]):

    @abc.abstractmethod
    def save(self, name: str, settings: T) -> int:
        """
        Parameters
        ----------
        name     : str
                   The unique name of the item to save.
                   The name cannot be all numbers.
        settings : T
                   The item to save.

        Raises
        ------
        ValueError
            If the item name already exists or is all numbers.

        Returns
        -------
        a unique identifier for the saved item.
        """

    @abc.abstractmethod
    def update(self, name: str | int, settings: T) -> RawSettings:
        """
        Parameters
        ----------
        name     : str | int
                   The unique name or int identifier of the item to update.
        settings : T
                   The item to update.

        Raises
        ------
        KeyError
            If the item does not exist.

        Returns
        -------
        The previous raw settings item for the given name.
        """

    @abc.abstractmethod
    def delete(self, name: str | int) -> RawSettings:
        """
        Parameters
        ----------
        name : str | int
               The unique name or int identifier of the settings to delete.

        Raises
        ------
        KeyError
            If the item does not exist.

        Returns
        -------
        The deleted raw settings item for the given name.
        """


class AsyncReadOnlyRegistry(abc.ABC, Generic[T]):

    @abc.abstractmethod
    async def ids(self, mode: Mode = "Valid") -> dict[int, str]:
        """
        Parameters
        ----------
        mode : Mode
               The mode to use when retrieving the ids.

        Returns
        -------
        A dictionary of the unique identifiers to the unique names.
        """

    @abc.abstractmethod
    async def names(self, mode: Mode = "Valid") -> dict[str, int]:
        """
        Parameters
        ----------
        mode : Mode
               The mode to use when retrieving the ids.

        Returns
        -------
        A dictionary of the unique names to the unique identifiers.
        """

    @abc.abstractmethod
    async def get_raw(self, name_or_id: list[str | int]) -> list[RawSettings]:
        """
        Parameters
        ----------
        name_or_id : list[str | int]
                     The unique names or int identifiers of the items to retrieve.

        Returns
        -------
        A list of `RawSettings` in the same order as input.
        """

    @abc.abstractmethod
    async def get(self, name: str | int) -> T:
        """
        Parameters
        ----------
        name : str | int
               The unique name or int identifier of the item to retrieve.

        Raises
        ------
        KeyError
            If the item does not exist.
        InvalidSettingsError:
            If the item exists but is invalid.

        Returns
        -------
        The item for the given name.
        """

    @abc.abstractmethod
    async def get_metadata(self, name: str | int) -> SettingsMetaData:
        """
        Parameters
        ----------
        name : str | int
               The unique name or int identifier of the item to retrieve.

        Raises
        ------
        KeyError
            If the item does not exist.

        Returns
        -------
        The metadata for the given name.
        """

    async def get_all(self) -> dict[str, T]:
        """
        Returns
        -------
        A dictionary of all valid available settings.
        """
        return {name: await self.get(name) for name in await self.names()}

    async def get_all_with_metadata(self) -> dict[str, tuple[T, SettingsMetaData]]:
        """
        Returns
        -------
        A dictionary of all valid available settings with metadata.
        """
        all_settings = await self.get_all()
        all_metadata = await self.get_all_metadata()

        return {
            name: (all_settings[name], all_metadata[name])
            for name in all_settings.keys()
        }

    async def get_all_metadata(self) -> dict[str, SettingsMetaData]:
        """
        Returns
        -------
        A dictionary of all available settings metadata, valid or invalid.
        """
        return {name: await self.get_metadata(name) for name in await self.names()}


class AsyncRegistry(Generic[T], AsyncReadOnlyRegistry[T]):

    @abc.abstractmethod
    async def save(self, name: str, settings: T) -> int:
        """
        Parameters
        ----------
        name     : str
                   The unique name of the item to save.
                   The name cannot be all numbers.
        settings : T
                   The item to save.

        Raises
        ------
        ValueError
            If the item name already exists or is all numbers.

        Returns
        -------
        a unique identifier for the saved item.
        """

    @abc.abstractmethod
    async def update(self, name: str | int, settings: T) -> RawSettings:
        """
        Parameters
        ----------
        name     : str | int
                   The unique name or int identifier of the item to update.
        settings : T
                   The item to update.

        Raises
        ------
        KeyError
            If the item does not exist.

        Returns
        -------
        The previous raw settings item for the given name.
        """

    @abc.abstractmethod
    async def delete(self, name: str | int) -> RawSettings:
        """
        Parameters
        ----------
        name : str | int
               The unique name or int identifier of the settings to delete.

        Raises
        ------
        KeyError
            If the item does not exist.

        Returns
        -------
        The deleted raw settings item for the given name.
        """


class SettingsRegistry(Registry[SettingsType], Generic[SettingsType, SettingsMenuType]):

    @abc.abstractmethod
    def available_settings(self) -> SettingsMenuType:
        """
        Returns
        -------
        A description of valid settings for this registry.
        """

    def save(self, name: str, settings: SettingsType) -> int:
        if re.sub(r"\d", "", name) == "":
            raise ValueError(
                f"The model model name cannot consist of only numbers: {name}",
            )
        self.available_settings().validate_settings(settings)
        return self._do_save(name, settings)

    @abc.abstractmethod
    def _do_save(self, name: str, settings: SettingsType) -> int: ...

    def update(self, name: str | int, settings: SettingsType) -> RawSettings:
        self.available_settings().validate_settings(settings)
        return self._do_update(name, settings)

    @abc.abstractmethod
    def _do_update(self, name: str | int, settings: SettingsType) -> RawSettings: ...


class AsyncSettingsRegistry(
    AsyncRegistry[SettingsType], Generic[SettingsType, SettingsMenuType]
):

    @abc.abstractmethod
    async def available_settings(self) -> SettingsMenuType:
        """
        Returns
        -------
        A description of valid settings for this registry.
        """

    async def save(self, name: str, settings: SettingsType) -> int:
        if re.sub(r"\d", "", name) == "":
            raise ValueError(
                f"The model model name cannot consist of only numbers: {name}",
            )
        if name in await self.names():
            raise ValueError(f"Name {name} already exists.")
        (await self.available_settings()).validate_settings(settings)
        return await self._do_save(name, settings)

    @abc.abstractmethod
    async def _do_save(self, name: str, settings: SettingsType) -> int: ...

    async def update(self, name: str | int, settings: SettingsType) -> RawSettings:
        (await self.available_settings()).validate_settings(settings)
        return await self._do_update(name, settings)

    @abc.abstractmethod
    async def _do_update(
        self, name: str | int, settings: SettingsType
    ) -> RawSettings: ...


class AsyncInMemorySettingsRegistry(
    AsyncSettingsRegistry[SettingsType, SettingsMenuType]
):

    def __init__(
        self,
        model_type: type[SettingsType],
        settings_menu: SettingsMenuType,
        resolver: AsyncSettingsResolver,
    ):
        self._model_type = model_type
        self._settings_menu = settings_menu
        self._resolver = resolver

        self._settings: dict[int, SettingsType] = {}
        self._metadata: dict[int, SettingsMetaData] = {}
        self._id_name_map: dict[int, str] = {}
        self._name_id_map: dict[str, int] = {}
        self._next_id = 0

    async def ids(self, mode: Mode = "Valid") -> dict[int, str]:
        all_ids = dict(self._id_name_map)
        if mode == "All":
            return all_ids
        elif mode in {"Valid", "Invalid"}:
            all_settings: list[Settings] = [self._settings[i] for i in all_ids]
            resolved = await self._resolver.resolve_settings(all_settings)
            for identifier, resolved_setting in zip(
                list(all_ids), resolved, strict=True
            ):
                valid = self._resolver.is_valid(resolved_setting)
                if mode == "Valid" and not valid:
                    all_ids.pop(identifier)
                elif mode == "Invalid" and valid:
                    all_ids.pop(identifier)
            return all_ids
        else:
            raise ValueError(f"Unknown mode: {mode}")

    async def names(self, mode: Mode = "Valid") -> dict[str, int]:
        all_names = dict(self._name_id_map)
        if mode == "All":
            return all_names
        elif mode in {"Valid", "Invalid"}:
            all_settings: list[Settings] = [
                self._settings[self._name_id_map[n]] for n in all_names
            ]
            resolved = await self._resolver.resolve_settings(all_settings)
            for name, resolved_setting in zip(list(all_names), resolved, strict=True):
                valid = self._resolver.is_valid(resolved_setting)
                if mode == "Valid" and not valid:
                    all_names.pop(name)
                elif mode == "Invalid" and valid:
                    all_names.pop(name)

            return all_names
        else:
            raise ValueError(f"Unknown mode: {mode}")

    async def available_settings(self) -> SettingsMenuType:
        return self._settings_menu

    async def get_raw(self, name_or_id: list[str | int]) -> list[RawSettings]:
        references: list[tuple[ModelType, str | int]] = [
            (self._model_type.__name__, e) for e in name_or_id
        ]
        if len(references) == 0:
            return []
        return await self._resolver.resolve_references(references)

    async def get(self, name: str | int) -> SettingsType:
        raw_settings = (
            await self._resolver.resolve_references([(self._model_type.__name__, name)])
        )[0]
        if not raw_settings.exists:
            raise KeyError(f"Could not find settings for input: {name}")
        if not self._resolver.is_valid(raw_settings):
            raise InvalidSettingsError(raw_settings.raw_json)
        model = self._model_type.model_validate(raw_settings.raw_json)
        return model

    async def get_metadata(self, name: str | int) -> SettingsMetaData:
        try:
            if isinstance(name, int):
                return self._metadata[name]
            return self._metadata[(await self.names())[name]]
        except KeyError as e:
            raise KeyError(f"Could not find settings for input: {name}") from e

    async def _do_save(self, name: str, settings: SettingsType) -> int:
        if name in self._name_id_map:
            raise ValueError(f"Settings with name {name!r} already exists.")
        self._settings[self._next_id] = settings
        self._metadata[self._next_id] = SettingsMetaData(
            created_on=dt.datetime.now(tz=dt.timezone.utc),
            last_updated=dt.datetime.now(tz=dt.timezone.utc),
        )
        self._id_name_map[self._next_id] = name
        self._name_id_map[name] = self._next_id
        self._next_id += 1
        return self._next_id - 1

    async def _do_update(self, name: str | int, settings: SettingsType) -> RawSettings:
        try:
            previous = (await self.get_raw([name]))[0]

            if isinstance(name, int):
                self._settings[name] = settings
                self._metadata[name] = self._metadata[name].model_copy(
                    update={"last_updated": dt.datetime.now(tz=dt.timezone.utc)}
                )
                return previous
            idx = (await self.names())[name]
            self._settings[idx] = settings
            self._metadata[idx] = self._metadata[idx].model_copy(
                update={"last_updated": dt.datetime.now(tz=dt.timezone.utc)}
            )
            return previous
        except KeyError as e:
            raise KeyError(f"Could not find settings for input: {name}") from e

    async def delete(self, name: str | int) -> RawSettings:
        try:
            raw_settings = (await self.get_raw([name]))[0]
            if not raw_settings.exists:
                raise KeyError()

            raw_settings = raw_settings.model_copy(update={"exists": False})

            if isinstance(name, int):
                actual_name = self._id_name_map.pop(name)
                self._name_id_map.pop(actual_name)
                self._metadata.pop(name)
                self._settings.pop(name)
                return raw_settings
            idx = (await self.names())[name]
            self._id_name_map.pop(idx)
            self._name_id_map.pop(name)
            self._metadata.pop(idx)
            self._settings.pop(idx)
            return raw_settings
        except KeyError as e:
            raise KeyError(f"Could not find settings for input: {name}") from e


ApiType = TypeVar("ApiType")


class RegistryBasedApi(abc.ABC, Generic[SettingsType, SettingsMenuType, ApiType]):

    @property
    @abc.abstractmethod
    def settings(self) -> SettingsRegistry[SettingsType, SettingsMenuType]: ...

    @abc.abstractmethod
    def load(
        self, ref_or_settings: str | int | SettingsType, *args: Any, **kwargs: Any
    ) -> ApiType: ...


class AsyncRegistryBasedApi(abc.ABC, Generic[SettingsType, SettingsMenuType, ApiType]):

    @property
    @abc.abstractmethod
    def settings(self) -> AsyncSettingsRegistry[SettingsType, SettingsMenuType]: ...

    @abc.abstractmethod
    async def load(
        self, ref_or_settings: str | int | SettingsType, *args: Any, **kwargs: Any
    ) -> ApiType: ...



import abc

from bayesline.api._src.equity.api import AsyncBayeslineEquityApi, BayeslineEquityApi
from bayesline.api._src.permissions import AsyncUserPermissionsApi, UserPermissionsApi


class BayeslineApi(abc.ABC):

    @property
    @abc.abstractmethod
    def equity(self) -> BayeslineEquityApi: ...

    @property
    @abc.abstractmethod
    def permissions(self) -> UserPermissionsApi: ...


class AsyncBayeslineApi(abc.ABC):

    @property
    @abc.abstractmethod
    def equity(self) -> AsyncBayeslineEquityApi: ...

    @property
    @abc.abstractmethod
    def permissions(self) -> AsyncUserPermissionsApi: ...



import datetime as dt
from typing import Any, Literal

import polars as pl

IdType = Literal[
    "bayesid",
    "ticker",
    "composite_figi",
    "cik",
    "cusip8",
    "cusip9",
    "isin",
    "sedol6",
    "sedol7",
    "proxy",
    "name",
]
DNFFilterExpression = tuple[str, str, Any]
DNFFilterExpressions = list[DNFFilterExpression | list[DNFFilterExpression]]

DateLike = str | dt.date | dt.datetime


def to_date(arg: DateLike) -> dt.date:
    """Cast a date-like object to a date.

    Parameters
    ----------
    arg : DateLike
        The date-like object to cast.

    Returns
    -------
    dt.date
        The date.
    """
    if isinstance(arg, dt.date):
        return arg
    elif isinstance(arg, dt.datetime):
        return arg.date()
    elif isinstance(arg, str):
        return pl.Series([arg]).str.to_date().to_list()[0]
    else:
        raise ValueError(f"Invalid date-like object: {arg}")


def to_maybe_date(arg: DateLike | None) -> dt.date | None:
    """Cast a date-like object, empty string or None to a date or None.

    Parameters
    ----------
    arg : DateLike | None
        The date-like object to cast.

    Returns
    -------
    dt.date | None
        The date or None.
    """
    if arg is None or arg == "":
        return None
    else:
        return to_date(arg)


def to_date_string(arg: DateLike) -> str:
    """Cast a date-like object to a date string.

    Parameters
    ----------
    arg : DateLike
        The date-like object to cast.

    Returns
    -------
    str
        The date string.
    """
    return to_date(arg).isoformat()


def to_maybe_date_string(arg: DateLike | None) -> str | None:
    """Cast a date-like object, empty string or None to a date string or None.

    Parameters
    ----------
    arg : DateLike | None
        The date-like object to cast.

    Returns
    -------
    str | None
        The date string or None.
    """
    if arg is None or arg == "":
        return None
    else:
        return to_date_string(arg)



import abc
from typing import Any


class AsyncUserPermissionsApi(abc.ABC):

    @abc.abstractmethod
    async def get_permissions_map(self) -> dict[str, Any]: ...

    @abc.abstractmethod
    async def get_perm(self, key: str, default: bool = True) -> Any: ...

    @abc.abstractmethod
    async def get_perms(
        self, keys: list[str], default: bool = True
    ) -> dict[str, Any]: ...


class UserPermissionsApi(abc.ABC):

    @abc.abstractmethod
    def get_permissions_map(self) -> dict[str, Any]: ...

    @abc.abstractmethod
    def get_perm(self, key: str, default: bool = True) -> Any: ...

    @abc.abstractmethod
    def get_perms(self, keys: list[str], default: bool = True) -> dict[str, Any]: ...



import os
from typing import Annotated, Literal

from pydantic import Field

from bayesline.api._src.registry import Settings, SettingsMenu, SettingsTypeMetaData


class PortfolioUploadSettings(Settings):
    """
    Specifies options for a portfolio upload, such as the parser etc.
    """

    parser: str = Field(description=("The parser to use."))
    portfolio_ids: list[str] = Field(
        description=("The available portfolio IDs that are contained in this upload.")
    )


class PortfolioUploadSettingsMenu(
    SettingsMenu[PortfolioUploadSettings], frozen=True, extra="forbid"
):

    parsers: list[str] = Field(
        description=(
            "The list of available parsers that can be used "
            "to parse the portfolio files."
        )
    )

    def describe(self, settings: PortfolioUploadSettings | None = None) -> str:
        if settings is None:
            return f"Parsers: [{', '.join(self.parsers)}]"
        else:
            all_ids = settings.portfolio_ids
            return (
                f"Parser: {settings.parser}{os.linesep}"
                f"Available Portfolios: {', '.join(all_ids)}"
            )

    def validate_settings(self, settings: PortfolioUploadSettings) -> None:
        if settings.parser not in self.parsers:
            raise ValueError(
                f"Invalid parser: {settings.parser}. "
                f"Available parsers are: {', '.join(self.parsers)}"
            )


class PortfolioOrganizerSettings(Settings):
    """
    Specifies which portfolios to enable (from different sources).

    Different sources (e.g. uploaded portfolios) can provide the same portfolio
    identifiers. These settings allow to specify which portfolios to enable from
    which sources.
    """

    enabled_portfolios: Annotated[
        dict[str, str],
        Field(
            description=(
                "The enabled portfolios from different sources. "
                "The key is the portfolio ID, and the value is the source "
                "(name of the underlying portfolio service)."
            ),
        ),
        SettingsTypeMetaData[dict[str, str]](
            references=PortfolioUploadSettings,
            extractor=lambda v: list(set(v.values())),
        ),
    ]


class PortfolioOrganizerSettingsMenu(
    SettingsMenu[PortfolioOrganizerSettings], frozen=True, extra="forbid"
):

    sources: dict[str, list[str]] = Field(
        description=(
            "Mapping of sources to the available portfolio IDs for that source."
        )
    )

    def describe(self, settings: PortfolioOrganizerSettings | None = None) -> str:
        if settings is None:
            return f"Sources: {self.sources}"
        else:
            enabled_portfolios = settings.enabled_portfolios
            return f"Enabled Portfolios: {enabled_portfolios}"

    def validate_settings(self, settings: PortfolioOrganizerSettings) -> None:
        messages = []
        for portfolio_id, source in settings.enabled_portfolios.items():
            if source not in self.sources:
                messages.append(
                    f"Invalid source: {source}. "
                    f"Available sources: {', '.join(self.sources.keys())}. "
                )
            elif portfolio_id not in self.sources[source]:
                messages.append(
                    f"Invalid portfolio ID: {portfolio_id} for source {source}. "
                )
        if messages:
            raise ValueError("".join(messages))


class PortfolioSettings(Settings):
    """
    Specifies different options of obtaining portfolios.
    """

    portfolio_schema: Annotated[
        str | int | PortfolioOrganizerSettings,
        Field(
            description=(
                "The portfolio organizer settings to use as an underlying schema of "
                "portfolios. The 'Default' schema is used by default."
            ),
        ),
        SettingsTypeMetaData[str | int | PortfolioOrganizerSettings](
            references=PortfolioOrganizerSettings
        ),
    ]
    ffill: Literal["no-ffill", "ffill-with-drift"] = "no-ffill"
    unpack: Literal["no-unpack", "unpack"] = "no-unpack"


class PortfolioSettingsMenu(
    SettingsMenu[PortfolioSettings], frozen=True, extra="forbid"
):
    """
    Specifies the set of available options that
    can be used to create portfolio settings.
    """

    schemas: list[str] = Field(
        description=(
            "The available schemas (i.e. names from the portfolio organizer)."
        ),
        default_factory=list,
    )

    def describe(self, settings: PortfolioSettings | None = None) -> str:
        if settings is None:
            return (
                f"Schemas: {', '.join(self.schemas)}\n"
                "Forward Fill Options: 'no-ffill', 'ffill-with-drift'\n"
                "Unpack Options: 'no-unpack', 'unpack"
            )
        else:
            return f"Forward Fill: {settings.ffill}\nUnpack: {settings.unpack}"

    def validate_settings(self, settings: PortfolioSettings) -> None:
        s = settings.portfolio_schema
        if isinstance(s, str) and s not in self.schemas:
            raise ValueError(
                f"Invalid schema: {settings.portfolio_schema}. "
                f"Available schemas are: {', '.join(self.schemas)}"
            )



import abc
import datetime as dt
from typing import Literal

import polars as pl
from pydantic import BaseModel

from bayesline.api._src.equity.portfolio_settings import (
    PortfolioOrganizerSettings,
    PortfolioOrganizerSettingsMenu,
    PortfolioSettings,
    PortfolioSettingsMenu,
    PortfolioUploadSettings,
    PortfolioUploadSettingsMenu,
)
from bayesline.api._src.registry import (
    AsyncRegistryBasedApi,
    AsyncSettingsRegistry,
    RawSettings,
    RegistryBasedApi,
    SettingsRegistry,
)
from bayesline.api._src.types import DateLike, IdType


class BayeslinePortfolioApi(abc.ABC):

    @property
    @abc.abstractmethod
    def name(self) -> str: ...

    @abc.abstractmethod
    def get_id_types(self) -> dict[str, list[IdType]]:
        """
        Returns
        -------
        dict[str, list[IdType]]:
            The available ID types that at least a portion of assets can be mapped
            to for each portfolio.
        """

    @abc.abstractmethod
    def get_coverage(
        self,
        names: str | list[str] | None = None,
        *,
        by: Literal["date", "asset"] = "date",
        metric: Literal["count", "holding"] = "count",
        stats: list[str] | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        names: str | list[str], optional
            The names of the portfolios.
        by: str, optional
            The coverage aggregation, either by date or by asset.
        metric: str, optional
            The metric to calculate, either count of observations
            or sum of holding values.
        stats: list[str], optional
            list of 'min', 'max', 'mean', collapses `by` into these stats.

        Returns
        -------
        pl.DataFrame:
            The dated coverage count for each id type. `portfolio_group` and
            `portfolio` are the first two columns. If `stats` given, collapses the `by`
            index to the given aggregations.
        """

    @abc.abstractmethod
    def get_portfolio_names(self) -> list[str]: ...

    @abc.abstractmethod
    def get_portfolio_groups(self) -> dict[str, list[str]]: ...

    @abc.abstractmethod
    def get_dates(
        self, names: list[str] | str | None = None, *, collapse: bool = False
    ) -> dict[str, list[dt.date]]: ...

    @abc.abstractmethod
    def get_portfolio(
        self,
        names: list[str] | str,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the portfolios for the given names between given start and end dates.

        Parameters
        ----------
        names: list[str] | str
            The list of portfolio names.
        start_date: DateLike, optional
            The start date of the data to return, inclusive.
        end_date: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            id type to return the portfolio holdings in.

        Returns
        -------
        pl.DataFrame:
            A dataframe with columns `portfolio_group`, `portfolio`, `date`,
            `input_asset_id`, `input_asset_id_type`, `asset_id`, `asset_id_type` and
            `value`.

            If no `id_type` is given then the input ID space will be used unmapped. In
            this case the columns `asset_id`, `asset_id_type` will not be returned.
        """


class AsyncBayeslinePortfolioApi(abc.ABC):

    @property
    @abc.abstractmethod
    def name(self) -> str: ...

    @abc.abstractmethod
    async def get_id_types(self) -> dict[str, list[IdType]]:
        """
        Returns
        -------
        dict[str, list[IdType]]:
            The available ID types that at least a portion of assets can be mapped
            to for each portfolio.
        """

    @abc.abstractmethod
    async def get_coverage(
        self,
        names: str | list[str] | None = None,
        *,
        by: Literal["date", "asset"] = "date",
        metric: Literal["count", "holding"] = "count",
        stats: list[str] | None = None,
    ) -> pl.DataFrame:
        """
        Parameters
        ----------
        names: str | list[str], optional
            The names of the portfolios. If not given all portfolios will be calculated.
        by: str, optional
            The coverage aggregation, either by date or by asset.
        metric: str, optional
            The metric to calculate, either count of observations
            or sum of holding values.
        stats: list[str], optional
            list of 'min', 'max', 'mean', collapses `by` into these stats.

        Returns
        -------
        pl.DataFrame:
            The dated coverage count for each id type. `portfolio_group` and
            `portfolio` are the first two columns. If `stats` given, collapses the `by`
            index to the given aggregations.
        """

    @abc.abstractmethod
    async def get_portfolio_names(self) -> list[str]: ...

    @abc.abstractmethod
    async def get_portfolio_groups(self) -> dict[str, list[str]]: ...

    @abc.abstractmethod
    async def get_dates(
        self, names: list[str] | str | None = None, *, collapse: bool = False
    ) -> dict[str, list[dt.date]]: ...

    @abc.abstractmethod
    async def get_portfolio(
        self,
        names: list[str] | str,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        id_type: IdType | None = None,
    ) -> pl.DataFrame:
        """
        Obtains the portfolios for the given names between given start and end dates.

        Parameters
        ----------
        names: list[str] | str
            The list of portfolio names.
        start_date: DateLike, optional
            The start date of the data to return, inclusive.
        end_date: DateLike, optional
            The end date of the data to return, inclusive.
        id_type: IdType, optional
            id type to return the portfolio holdings in.

        Returns
        -------
        pl.DataFrame:
            A dataframe with columns `portfolio_group`, `portfolio`, `date`,
            `input_asset_id`, `input_asset_id_type`, `asset_id`, `asset_id_type` and
            `value`.

            If no `id_type` is given then the input ID space will be used unmapped. In
            this case the columns `asset_id`, `asset_id_type` will not be returned.
        """


class PortfolioParserResult(BaseModel, frozen=True, extra="forbid"):

    parser: str | None = None
    success: bool
    messages: list[str]
    portfolio_ids: list[str]


class PortfolioParser(abc.ABC):

    @property
    @abc.abstractmethod
    def name(self) -> str: ...

    @abc.abstractmethod
    def get_examples(self) -> list[pl.DataFrame]:
        """
        Returns
        -------
        list[pl.DataFrame]:
            A list of example portfolios that can be used to test the parser.
        """

    @abc.abstractmethod
    def can_handle(self, df: pl.DataFrame) -> PortfolioParserResult:
        """
        Parameters
        ----------
        df: pl.DataFrame
            The dataframe to check if the parser can handle.

        Returns
        -------
        PortfolioParserResult:
            The result of the parser check.
        """

    @abc.abstractmethod
    def parse(self, df: pl.DataFrame) -> tuple[pl.DataFrame, PortfolioParserResult]:
        """
        Parameters
        ----------
        df: pl.DataFrame
            The dataframe to parse.

        Returns
        -------
        a tuple of pl.DataFrame and PortfolioParserResult:
            if the parser was unsuccessful the parsed
            dataframe will be an empty dataframe.
        """


class AsyncPortfolioParser(abc.ABC):

    @property
    @abc.abstractmethod
    def name(self) -> str: ...

    @abc.abstractmethod
    async def get_examples(self) -> list[pl.DataFrame]:
        """
        Returns
        -------
        list[pl.DataFrame]:
            A list of example portfolios that can be used to test the parser.
        """

    @abc.abstractmethod
    async def can_handle(self, df: pl.DataFrame) -> PortfolioParserResult:
        """
        Parameters
        ----------
        df: pl.DataFrame
            The dataframe to check if the parser can handle.

        Returns
        -------
        PortfolioParserResult:
            The result of the parser check.
        """

    @abc.abstractmethod
    async def parse(
        self, df: pl.DataFrame
    ) -> tuple[pl.DataFrame, PortfolioParserResult]:
        """
        Parameters
        ----------
        df: pl.DataFrame
            The dataframe to parse.

        Returns
        -------
        a tuple of pl.DataFrame and PortfolioParserResult:
            if the parser was unsuccessful the parsed
            dataframe will be an empty dataframe.
        """


class BayeslinePortfolioUploadApi(abc.ABC):

    @abc.abstractmethod
    def get_parser_result(self) -> PortfolioParserResult: ...

    @abc.abstractmethod
    def get_raw(self) -> pl.DataFrame:
        """
        Returns
        -------
        pl.DataFrame:
            The raw data frame as it was uploaded.
        """

    @abc.abstractmethod
    def get_parsed(self) -> pl.DataFrame:
        """
        Returns
        -------
        pl.DataFrame:
            The parsed data frame.
        """


class AsyncBayeslinePortfolioUploadApi(abc.ABC):

    @abc.abstractmethod
    async def get_parser_result(self) -> PortfolioParserResult: ...

    @abc.abstractmethod
    async def get_raw(self) -> pl.DataFrame:
        """
        Returns
        -------
        pl.DataFrame:
            The raw data frame as it was uploaded.
        """

    @abc.abstractmethod
    async def get_parsed(self) -> pl.DataFrame:
        """
        Returns
        -------
        pl.DataFrame:
            The parsed data frame.
        """


class PortfolioUploadError(Exception):
    pass


class BayeslineEquityPortfolioUploadApi(
    RegistryBasedApi[
        PortfolioUploadSettings,
        PortfolioUploadSettingsMenu,
        BayeslinePortfolioUploadApi,
    ]
):
    @abc.abstractmethod
    def get_parser_names(self) -> list[str]:
        """
        Returns
        -------
        list[str]:
            The list of available parsers that can be used to parse the portfolio files.
        """

    @abc.abstractmethod
    def get_parser(self, parser: str) -> PortfolioParser:
        """
        Parameters
        ----------
        parser: str
            The parser to obtain.

        Returns
        -------
        PortfolioParser:
            The parser.
        """

    @abc.abstractmethod
    def can_handle(
        self, df: pl.DataFrame, *, parser: str | None = None
    ) -> PortfolioParserResult:
        """
        Parameters
        ----------
        df: pl.DataFrame
            The dataframe to check if the parser can handle.
        parser: str, optional
            The parser to use. If None will check all parsers.

        Returns
        -------
        PortfolioParserResult:
            The result of the parser check.
        """

    @abc.abstractmethod
    def upload(
        self, name: str, df: pl.DataFrame, *, parser: str | None = None
    ) -> PortfolioUploadSettings:
        """
        Parameters
        ----------
        name: str
            The name of the upload.
        df: pl.DataFrame
            The dataframe to upload.
        parser: str, optional
            The parser to use. If None will check all parsers.

        Raises
        ------
            PortfolioUploadError: If the upload fails.

        Returns
        -------
        PortfolioUploadSettings:
            The settings for the upload.
        """

    @abc.abstractmethod
    def delete(self, name: str) -> RawSettings:
        """
        Parameters
        ----------
        name: str
            The name of the upload to delete.

        Returns
        -------
        RawSettings:
            The raw settings for the deleted upload.
        """


class AsyncBayeslineEquityPortfolioUploadApi(
    AsyncRegistryBasedApi[
        PortfolioUploadSettings,
        PortfolioUploadSettingsMenu,
        AsyncBayeslinePortfolioUploadApi,
    ]
):
    @abc.abstractmethod
    async def get_parser_names(self) -> list[str]:
        """
        Returns
        -------
        list[str]:
            The list of available parsers that can be used to parse the portfolio files.
        """

    @abc.abstractmethod
    async def get_parser(self, parser: str) -> AsyncPortfolioParser:
        """
        Parameters
        ----------
        parser: str
            The parser to obtain.

        Returns
        -------
        PortfolioParser:
            The parser.
        """

    @abc.abstractmethod
    async def can_handle(
        self, df: pl.DataFrame, *, parser: str | None = None
    ) -> PortfolioParserResult:
        """
        Parameters
        ----------
        df: pl.DataFrame
            The dataframe to check if the parser can handle.
        parser: str, optional
            The parser to use. If None will check all parsers.

        Returns
        -------
        PortfolioParserResult:
            The result of the parser check.
        """

    @abc.abstractmethod
    async def upload(
        self, name: str, df: pl.DataFrame, *, parser: str | None = None
    ) -> PortfolioUploadSettings:
        """
        Parameters
        ----------
        name: str
            The name of the upload.
        df: pl.DataFrame
            The dataframe to upload.
        parser: str, optional
            The parser to use. If None will check all parsers.

        Raises
        ------
            PortfolioUploadError: If the upload fails.

        Returns
        -------
        PortfolioUploadSettings:
            The settings for the upload.
        """

    @abc.abstractmethod
    async def delete(self, name: str) -> RawSettings:
        """
        Parameters
        ----------
        name: str
            The name of the upload to delete.

        Returns
        -------
        RawSettings:
            The raw settings for the deleted upload.
        """


class BayeslineEquityPortfolioApi(
    RegistryBasedApi[
        PortfolioSettings,
        PortfolioSettingsMenu,
        BayeslinePortfolioApi,
    ]
):
    @property
    @abc.abstractmethod
    def uploader(self) -> BayeslineEquityPortfolioUploadApi: ...

    @property
    @abc.abstractmethod
    def organizer_settings(
        self,
    ) -> SettingsRegistry[
        PortfolioOrganizerSettings, PortfolioOrganizerSettingsMenu
    ]: ...


class AsyncBayeslineEquityPortfolioApi(
    AsyncRegistryBasedApi[
        PortfolioSettings,
        PortfolioSettingsMenu,
        AsyncBayeslinePortfolioApi,
    ]
):
    @property
    @abc.abstractmethod
    def uploader(self) -> AsyncBayeslineEquityPortfolioUploadApi: ...

    @property
    @abc.abstractmethod
    def organizer_settings(
        self,
    ) -> AsyncSettingsRegistry[
        PortfolioOrganizerSettings, PortfolioOrganizerSettingsMenu
    ]: ...
